{ KEYWORDS: VERSION }
{ VERSION: 1.0 }
{ 1.0 / 29.10.2005
  - nova funkce SetRoot() umoznuje prepnout na jiny lokalizacni adresar
  0.6 / 18.9.2005
  - znovuoziveni
  - opraveno: sloka se soubory se vytvo jenom v reimu BUILDEM  
  0.5 / 21.11.2000
  - opravena chyba v prilisnem rozteceni rekurze v MapComponent() -> mapper.AddComponent()
  0.4
  - GetLanguage/SetLanguage prepinaji soubor s textama a aktualizuji komponenty
    na ktere byl pouzit MapComponent()
  0.3
  - pridany fce LoadMsgDef a LoadStringDef, ktere berou jako parametr implicitni
    hodnotu. v rezimu BUILDEM aktualizuji databazi
}
unit DXResource;

interface
uses
  TypInfo, Windows, Graphics, Classes;

procedure SetRoot(const Folder: string);

procedure SetLanguage(blabla: string);
function GetLanguage: string;

function LoadBitmap(const key: string): TBitmap;

function LoadString(const key: string): string;
function LoadMsg(key: integer): string;

function LoadStringDef(const key: string; const defaultValue: string): string;
function LoadMsgDef(key: integer; const defaultValue: string): string;

procedure AssignGlyph(Owner: TObject; const key: string);
procedure AssignText(Owner: TComponent; const key: string);

procedure AssignBitmapProperty(Owner: TObject; const propertyName: string; const key: string);
procedure AssignTextProperty(Owner: TObject; const propertyName: string; const key: string);

{ napr. HasProperty(pepa,'Text',[tkString,tkLString]) -> true kdyz objekt pepa ma vlastnost Text typu string
        HasProperty(pepa,'Nazdar',[tkInteger]) -> true kdyz pepa ma vlastnost Nazdar typu cele cislo }
function HasProperty(Owner: TObject; const propertyName: string; propertyKinds: TTypeKinds): boolean;

const
  prString  = 0;
  prBitmap  = 1;

procedure MapComponent(
  Owner: TComponent;               { soucastka, ktera se ma nasat }
  const key: string;               { jmeno, pod kterym jsou jeji udaje v databazi }
  const properties: array of const { seznam vlastnosti, ktere se maji prebouchat vzdycky v parech
                                     pr.. konstanta a jmeno vlastnosti
                                     napr. [prString,'Caption',prString,'Hint']
                                   }
  );

implementation
uses
  SysUtils;

{ TUNE: }
{ $DEFINE EXTERNALRES}
{ $DEFINE BUILDEM}

function R2S(const s: string): string;
var
  i: integer;
begin
  Result := '';
  i := 1;
  while i <= length(s) do begin
    if s[i] = '\' then begin
      inc(i);
      case s[i] of
        'n': Result := Result + #13#10;
        't': Result := Result + #9;
        'r': Result := Result + '=';
      else
        Result := Result + s[i]
      end
    end
    else
      Result := Result + s[i];
    inc(i)
  end
end;

function S2R(const s: string): string;
var
  i: integer;
begin
  Result := '';
  for i := 1 to length(s) do
    case s[i] of
      #10: ;
      #13: Result := Result + '\n';
      #9: Result := Result + '\t';
      '=': Result := Result + '\r';
      '\': Result := Result + '\\';
    else
      Result := Result + s[i]
    end
end;

function HasProperty(Owner: TObject; const propertyName: string; propertyKinds: TTypeKinds): boolean;
var
  info: PPropInfo;
begin
  info := GetPropInfo(Owner.ClassInfo, propertyName);
  if info = nil then
    Result := false
  else
    Result := info^.PropType^.Kind in propertyKinds
end;

type
  TNameStrings = class
  private
    FCount: Integer;
    FNames: TStringList;
    FValues: TStringList;
    function GetName(Index: Integer): string;
    function GetValueFromIndex(Index: Integer): string;
    function GetValue(const Name: string): string;
    procedure SetValue(const Name: string; const Value: string);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Source: TStrings);
    procedure LoadFromStream(Stream: TStream);
    procedure LoadFromFile(const FileName: string);
    procedure SaveToFile(const FileName: string);
    function IndexOfName(const Name: string): Integer;
    property Names[Index: Integer]: string read GetName;
    property ValueFromIndex[Index: Integer]: string read GetValueFromIndex;
    property Values[const Name: string]: string read GetValue write SetValue;
    property Count: Integer read FCount;
  end;

{$IFDEF EXTERNALRES}
var
  Lib: THandle;
{$ELSE}
var
  Root: string;
  // StringTableFileName:String{'_STRINGS.TXT'};
{$ENDIF}
var
  Strings: TNameStrings;
  wasError: boolean;

const
  StringTableName = '_STRINGS';

var
  StringTableFileName: string;

const
  NoDefault = #1;

function LoadMsg(key: integer): string;
begin
  Result := LoadMsgDef(key, NoDefault)
end;

function InternalLoadString(const key: string; const defaultValue: string): string;
{$IFDEF EXTERNALRES}
var
  mem: TResourceStream;
{$ENDIF}
begin
  if not Assigned(Strings) then begin
    Strings := TNameStrings.Create;
{$IFDEF EXTERNALRES}
    try
      mem := TResourceStream.Create(Lib, StringTableName, RT_RCDATA);
      try
        Strings.LoadFromStream(mem)
      finally
        mem.Free
      end
    except
    end
{$ELSE}
    if FileExists(Root + StringTableFilename) then
    try
      Strings.LoadFromFile(Root + StringTableFilename)
    except
    end
{$ENDIF}
  end;
  wasError := Strings.IndexOfName(key) < 0;
  if wasError and (defaultValue <> NoDefault) then begin
{$IFDEF BUILDEM}
    Strings.Values[key] := S2R(defaultValue);
{$ENDIF}
    Result := defaultValue
  end
  else
    Result := R2S(Strings.Values[key]);
end;

function LoadString(const key: string): string;
begin
  Result := InternalLoadString(key, NoDefault)
end;

function LoadStringDef(const key: string; const defaultValue: string): string;
begin
  Result := InternalLoadString(key, defaultValue)
end;

function LoadMsgDef(key: integer; const defaultValue: string): string;
begin
  Result := InternalLoadString('#' + IntToStr(key), defaultValue)
end;

function LoadBitmap(const key: string): TBitmap;
begin
  Result := TBitmap.Create;
  try
{$IFDEF EXTERNALRES}
    Result.LoadFromResourceName(Lib, UpperCase(key));
{$ELSE}
    Result.LoadFromFile(Root + key + '.bmp')
{$ENDIF}
  except
    Result.Free;
    Result := nil
  end
end;

procedure AssignGlyph(Owner: TObject; const key: string);
begin
  AssignBitmapProperty(Owner, 'Glyph', key)
end;

procedure AssignText(Owner: TComponent; const key: string);
begin
  AssignTextProperty(Owner, key, Owner.Name)
end;

procedure AssignBitmapProperty(Owner: TObject; const propertyName: string; const key: string);
var
  info: PPropInfo;
  b: TBitmap;
begin
  b := LoadBitmap(key);
  if Assigned(b) then begin
    try
      info := GetPropInfo(Owner.ClassInfo, propertyName);
      if (info <> nil) and (info^.PropType^.Kind = tkClass) then
        SetOrdProp(Owner, info, integer(pointer(b)))
    finally
      b.Free
    end
  end
end;

procedure AssignTextProperty(Owner: TObject; const propertyName: string; const key: string);
var
  info: PPropInfo;
begin
  info := GetPropInfo(Owner.ClassInfo, propertyName);
  if (info = nil) or not (info^.PropType^.Kind in [tkString, tkLString]) then
    exit;
  SetStrProp(Owner, info, LoadStringDef(key + '.' + propertyName, GetStrProp(Owner, info)));
end;

type
  TMapper = class(TComponent)
  private
    FList: TStrings;
    FProperties: TStrings;
  public
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure AddComponent(
      Owner: TComponent;               { soucastka, ktera se ma nasat }
      const key: string;               { jmeno, pod kterym jsou jeji udaje v databazi }
      const properties: array of const { seznam vlastnosti, ktere se maji prebouchat vzdycky v parech
                                         pr.. konstanta a jmeno vlastnosti
                                         napr. [prString,'Caption',prString,'Hint']
                                       }
      );
    procedure RemapComponents;
    function Registered(AComponent: TComponent): boolean;
    constructor Create;
    destructor Destroy; override;
  end;

constructor TMapper.Create;
begin
  inherited Create(nil);
  FList := TStringList.Create;
  FProperties := TStringList.Create
end;

destructor TMapper.Destroy;
begin
  inherited;
  FList.Free;
  FProperties.Free
end;

function TMapper.Registered(AComponent: TComponent): boolean;
begin
  Result := FList.IndexOfObject(AComponent) >= 0
end;

procedure TMapper.Notification(AComponent: TComponent; Operation: TOperation);
var
  i: integer;
begin
  inherited;
  if Operation = opRemove then begin
    i := FList.IndexOfObject(AComponent);
    if i >= 0 then begin
      FList.Delete(i);
      FProperties.Delete(i)
    end
  end;
end;

procedure TMapper.AddComponent(Owner: TComponent; const key: string; const properties: array of const);
begin
  if Registered(Owner) then
    exit;
  FProperties.Add('?'); // TODO:
  FList.AddObject(key, Owner);
  Owner.FreeNotification(Self)
end;

procedure TMapper.RemapComponents;
var
  i: integer;
begin
  i := 0;
  while i < FList.Count do begin
    MapComponent(TComponent(FList.Objects[i]), FList[i], [prString, 'Hint', prString, 'Caption', prString, 'Title', prString, 'Filter']);
    inc(i)
  end
end;

var
  mapper: TMapper;

procedure MapComponent(Owner: TComponent; const key: string; const properties: array of const);

  procedure InternalMapComponent(Owner: TComponent; const key: string);
  var
    i: integer;
    j: integer;
    r: integer;
    propertyName: string;
    c: TComponent;
  begin
    j := 0;
    while j < High(properties) do begin
      r := properties[j].VInteger;
      case properties[j + 1].VType of
        vtString: propertyName := properties[j + 1].VString^;
        vtAnsiString: propertyName := string(properties[j + 1].VAnsiString);
      else
        assert(false)
      end;
      inc(j, 2);
      case r of
        prString:
          if HasProperty(Owner, propertyName, [tkString, tkLString]) then begin
            AssignTextProperty(Owner, propertyName, key);
          end;
        prBitmap:
          if HasProperty(Owner, propertyName, [tkClass]) then begin
            AssignBitmapProperty(Owner, propertyName, key + '.' + propertyName);
          end
      end
    end;
    for i := 0 to Owner.ComponentCount - 1 do begin
      c := Owner.Components[i];
      if (c.Name = '') or mapper.Registered(c) then
        continue;
      InternalMapComponent(c, key + '.' + c.Name)
    end
  end;

begin
  if Assigned(mapper) then
  try
    InternalMapComponent(Owner, key);
  finally
    mapper.AddComponent(Owner, key, properties);
  end
end;

function GetLanguage: string;
begin
  Result := ChangeFileExt(ExtractFileName(StringTableFileName), '')
end;

procedure SaveStrings;
begin
  {$IFDEF BUILDEM}
  if Assigned(Strings) then begin
    CreateDir(Copy(Root, 1, length(Root) - 1));
    try
      Strings.SaveToFile(Root + StringTableFilename)
    except
    end
  end;
  {$ENDIF}
end;

procedure SetLanguage(blabla: string);
begin
  if blabla = ChangeFileExt(ExtractFileName(StringTableFileName), '') then
    exit;

  SaveStrings;
  Strings.Free;
  Strings := nil;

  StringTableFileName := blabla + '.lng';

  if not FileExists(Root + StringTableFileName) then
    StringTableFileName := 'en.lng';

  if Assigned(mapper) then
    mapper.RemapComponents
end;

function DefaultLanguage: string;
var
  LocaleName: array [0..4] of Char;
begin
  LocaleName[0] := #0;
  GetLocaleInfo(GetThreadLocale, LOCALE_SABBREVLANGNAME, LocaleName, sizeof(LocaleName));
  Result := Copy(StrPas(LocaleName), 1, 2)
end;

//-----------------------------------------------------------------------------

function TNameStrings.GetName(Index: Integer): string;
begin
  Result := FNames[Index]
end;

function TNameStrings.IndexOfName(const Name: string): Integer;
begin
  Result := FNames.IndexOf(Name)
end;

function TNameStrings.GetValueFromIndex(Index: Integer): string;
begin
  Result := FValues[Integer(FNames.Objects[Index])]
end;

function TNameStrings.GetValue(const Name: string): string;
var
  Index: Integer;
begin
  Index := FNames.IndexOf(Name);
  if Index < 0 then
    Result := ''
  else
    Result := FValues[Integer(FNames.Objects[Index])]
end;

procedure TNameStrings.SetValue(const Name: string; const Value: string);
var
  Index: Integer;
begin
  Index := FNames.IndexOf(Name);
  if Index < 0 then begin
    Index := FNames.Add(Name);
    FNames.Objects[Index] := TObject(FValues.Add(Value));
    Inc(FCount)
  end else
    FValues[Integer(FNames.Objects[Index])] := Value
end;

constructor TNameStrings.Create;
begin
  inherited Create;
  FNames := TStringList.Create;
  FNames.Sorted := True;
  FValues := TStringList.Create;
  FValues.Duplicates := dupAccept;
  FValues.Sorted := False
end;

procedure TNameStrings.Assign(Source: TStrings);
var
  I: Integer;
begin
  FNames.Clear;
  FValues.Clear;

  for I := 0 to Source.Count - 1 do
    FNames.AddObject(Source.Names[I], TObject(FValues.Add(Source.ValueFromIndex[I])));

  FCount := FNames.Count;
end;

destructor TNameStrings.Destroy;
begin
  FNames.Free;
  FValues.Free;
  inherited
end;

procedure TNameStrings.LoadFromStream(Stream: TStream);
var
  T: TStrings;
begin
  T := TStringList.Create;
  try
    T.LoadFromStream(Stream);
    Assign(T)
  finally
    T.Free
  end
end;

procedure TNameStrings.LoadFromFile(const FileName: string);
var
  T: TStrings;
begin
  T := TStringList.Create;
  try
    T.LoadFromFile(FileName);
    Assign(T)
  finally
    T.Free
  end
end;

procedure TNameStrings.SaveToFile(const FileName: string);
var
  Stream: TStream;
  I: Integer;
  S: string;
begin
  Stream := TFileStream.Create(FileName, fmCreate);
  try
    for I := 0 to Count - 1 do begin
      S := Names[I] + '=' + ValueFromIndex[I] + #13#10;
      Stream.WriteBuffer(S[1], Length(S));
    end
  finally
    Stream.Free
  end
end;

procedure SetRoot(const Folder: string);
begin
  Root := Folder;
  if (Root <> '') and (Root[Length(Root)] <> '\') then
    Root := Root + '\';
  StringTableFileName := '';
  SetLanguage(DefaultLanguage)
end;

//--------------------------------------------------------------

initialization
  Strings := nil;
  mapper := TMapper.Create;
{$IFDEF EXTERNALRES}
  Lib := LoadLibraryEx(PChar(ChangeFileExt(ParamStr(0), '.LNG')), 0, LOAD_LIBRARY_AS_DATAFILE {0});
{$ELSE}
  Root := ExtractFilePath(ParamStr(0)) + 'Lang\';
  SetLanguage(DefaultLanguage);
{$ENDIF}
finalization
  SaveStrings;
  Strings.Free;
{$IFDEF EXTERNALRES}
  FreeLibrary(Lib);
{$ENDIF}
  mapper.Free;
end.

