unit LRWPApp;

interface
uses Classes,WinShoes,HTTPApp;

 type
  TXitamiRequest=class(TWebRequest)
  private
   FContent:string;
  protected
   environment:TStrings;
   fromServer:TStream;
   toServer:TStream;
  public
   constructor Create(AnEnvironment:TStrings; AFromServer:TStream; AToServer:TStream);
  protected
   function GetStringVariable(Index: Integer): string; override;
   function GetDateVariable(Index: Integer): TDateTime; override;
   function GetIntegerVariable(Index: Integer): Integer; override;
  public
   // Read count bytes from client
// TODO:   function ReadClient(var Buffer; Count: Integer): Integer; override;
   // Read count characters as a string from client
   function ReadString(Count: Integer): string; override;
   // Translate a relative URI to a local absolute path
   function TranslateURI(const URI: string): string; override;
   // Write count bytes back to client
   function WriteClient(var Buffer; Count: Integer): Integer; override;
   // Write string contents back to client
   function WriteString(const AString: string): Boolean; override;
   // Read an arbitrary HTTP/Server Field not lists here
   function GetFieldByName(const Name: string): string; override;
  end;

  TXitamiResponse=class(TWebResponse)
  protected
   FSent:boolean;
   FStatusCode:integer;
   FStringVariables: array[0..MAX_STRINGS - 1] of string;
   FIntegerVariables: array[0..MAX_INTEGERS - 1] of Integer;
   FDateVariables: array[0..MAX_DATETIMES - 1] of TDateTime;
   FContent:string;
  protected
   function GetStringVariable(Index: Integer): string; override;
   procedure SetStringVariable(Index: Integer; const Value: string); override;
   function GetDateVariable(Index: Integer): TDateTime; override;
   procedure SetDateVariable(Index: Integer; const Value: TDateTime); override;
   function GetIntegerVariable(Index: Integer): Integer; override;
   procedure SetIntegerVariable(Index: Integer; Value: Integer); override;
   function GetContent: string; override;
   procedure SetContent(const Value: string); override;
   function GetStatusCode: Integer; override;
   procedure SetStatusCode(Value: Integer); override;
   function GetLogMessage: string; override;
   procedure SetLogMessage(const Value: string); override;
   function Sent:boolean;override;
  public
   constructor Create(HTTPRequest:TWebRequest);
   procedure SendResponse; override;
   procedure SendRedirect(const URI: string); override;
   procedure SendStream(AStream: TStream); override;
  end;

  TLRWP=class(TComponent)
  private
   procedure EmptyCaches;
  protected
   port:TWinShoeClient;
  public
   constructor Create(AnOwner:TComponent);override;
   destructor Destroy;override;
  public
   environment:TStrings;
   post:TStream;
   data:TStream;
  public
   procedure Connect(const AnAppname:string; const AServer:string; const APort:string; const Ahost:string);
   function AcceptRequest:boolean;
   procedure FinishRequest;
   procedure Disconnect;
  end;

  TXitamiApplication=class(TWebApplication)
  private
   worker:TThread;
   procedure OnWorkerSignal(Sender:TObject);
  protected
   FServer:string;
   Fhost:string;
   port:TLRWP;
   procedure Terminate;virtual;
   procedure ConnectServer;
   procedure ProcessMessages;virtual;
  public
   procedure Run;override;
  public
   property Server:string read FServer write FServer;
   property host:string read Fhost write Fhost;
  end;

implementation
uses Windows,Messages,SysUtils;

const LSeparator=#$FF;

procedure Error; begin raise Exception.Create('') end;

constructor TLRWP.Create;
begin
 inherited;
 environment:=TStringList.Create;
 post:=TMemoryStream.Create;
 data:=TMemoryStream.Create
end;

destructor TLRWP.Destroy;
begin
 environment.Free;
 post.Free;
 data.Free;
 inherited
end;

procedure TLRWP.Connect;
begin
 if not Assigned(port) then port:=TWinShoeClient.Create(Self);
 port.Disconnect;
 port.Host:=AServer;
 port.Port:=StrToInt(APort);
 port.Connect;
 port.Write(AnAppname+LSeparator+Ahost+LSeparator+#0#0);
 if UpperCase(Copy(port.ReadBuffer,1,2))<>'OK' then Error // TODO:
end;

procedure TLRWP.Disconnect;
begin
 port.Disconnect;
 EmptyCaches;
end;

procedure TLRWP.EmptyCaches;
begin
 post.Size:=0;
 post.Position:=0;
 data.Size:=0;
 data.Position:=0
end;

procedure TLRWP.FinishRequest;
var id:string;
begin
 try
  id:=IntToStr(data.Size);
  while length(id)<9 do id:='0'+id;
  port.Write(id);
  if data.Size>0 then port.WriteFromBuffer(TMemoryStream(data).Memory^,data.Size)
 finally
  EmptyCaches
 end
end;

function TLRWP.AcceptRequest;

 function ReadSize:integer;
 begin
  Result:=StrToInt(port.Read(9))
 end;

 function ZToText(const s:string):string;
 var i:integer;
 begin
  Result:='';
  for i:=1 to length(s) do
   if s[i]=#0 then
    Result:=Result+#13#10
   else
    Result:=Result+s[i]
 end;

var buf:string;
var envSize:integer;
begin
 try
  envSize:=ReadSize;
  environment.Clear;
  SetLength(buf,envSize);
  if envSize>0 then port.ReadToBuffer(buf[1],envSize);
  environment.Text:=ZToText(buf);

  post.Size:=ReadSize;
  if post.Size<>0 then begin
   port.ReadToBuffer(TMemoryStream(post).Memory^,post.Size);
   post.Position:=0
  end;

  Result:=true
 except
  Result:=false
 end
end;

//-----------------------------------------------------------------------------

constructor TXitamiRequest.Create;
begin
 environment:=AnEnvironment;
 fromServer:=AFromServer;
 toServer:=AToServer;
 inherited Create
end;

const
  CGIServerVariables: array[0..28] of string = (
    'REQUEST_METHOD',
    'SERVER_PROTOCOL',
    'URL',
    'QUERY_STRING',
    'PATH_INFO',
    'PATH_TRANSLATED',
    'HTTP_CACHE_CONTROL',
    'HTTP_DATE',
    'HTTP_ACCEPT',
    'HTTP_FROM',
    'HTTP_HOST',
    'HTTP_IF_MODIFIED_SINCE',
    'HTTP_REFERER',
    'HTTP_USER_AGENT',
    'HTTP_CONTENT_ENCODING',
    'HTTP_CONTENT_TYPE',
    'HTTP_CONTENT_LENGTH',
    'HTTP_CONTENT_VERSION',
    'HTTP_DERIVED_FROM',
    'HTTP_EXPIRES',
    'HTTP_TITLE',
    'REMOTE_ADDR',
    'REMOTE_HOST',
    'SCRIPT_NAME',
    'SERVER_PORT',
    '',
    'HTTP_CONNECTION',
    'HTTP_COOKIE',
    'HTTP_AUTHORIZATION');

function TXitamiRequest.GetStringVariable(Index: Integer): string;
begin
  if Index = 25 then
    Result := FContent
  else Result := GetFieldByName(CGIServerVariables[Index]);
end;

function TXitamiRequest.GetDateVariable(Index: Integer): TDateTime;
var
  Value: string;
begin
  Value := GetStringVariable(Index);
  if Value <> '' then
    Result := ParseDate(Value)
  else Result := -1;
end;

function TXitamiRequest.GetIntegerVariable(Index: Integer): Integer;
var
  Value: string;
begin
  Value := GetStringVariable(Index);
  Result := StrToIntDef(Value, -1)
end;

function TXitamiRequest.ReadString;
begin
 SetLength(Result,Count);
 if Count>0 then ReadClient(Result[1],Count)
end;

function TXitamiRequest.TranslateURI;
begin
 Result:=URI // TODO:
end;

function TXitamiRequest.WriteClient(var Buffer; Count: Integer): Integer;
begin
 Result:=toServer.Write(Buffer,Count)
end;

function TXitamiRequest.WriteString;
var L:integer;
begin
 L:=Length(AString);
 Result:=(L=0) or (WriteClient(pointer(@AString[1])^,L)=L)
end;

function TXitamiRequest.GetFieldByName(const Name: string): string;
var
  Buffer: array[0..4095] of Char;

  function StripHTTP(const Name: string): string;
  begin
    if Pos('HTTP_', Name) = 1 then
      Result := Copy(Name, 6, MaxInt)
    else Result := Name;
  end;

begin
  Result:=environment.Values[Name];
  if Result = '' then Result:=environment.Values[StripHTTP(Name)]
end;

//-----------------------------------------------------------------------------
constructor TXitamiResponse.Create;
begin
  inherited Create(HTTPRequest);
  if FHTTPRequest.ProtocolVersion = '' then
    Version := '1.0';
  StatusCode := 200;
  LastModified := -1;
  Expires := -1;
  Date := -1;
  ContentType := 'text/html';
end;

function TXitamiResponse.GetStringVariable;
begin
  if (Index >= Low(FStringVariables)) and (Index <= High(FStringVariables)) then
    Result := FStringVariables[Index]
  else
    Result:= ''
end;

procedure TXitamiResponse.SetStringVariable;
begin
 if (Index>=Low(FStringVariables)) and (Index<=High(FStringVariables)) then
  FStringVariables[Index]:=Value
end;

function TXitamiResponse.GetDateVariable;
begin
 if (Index>=Low(FDateVariables)) and (Index<=High(FDateVariables)) then
  Result:=FDateVariables[Index]
 else
  Result:=0
end;

procedure TXitamiResponse.SetDateVariable;
begin
 if (Index>=Low(FDateVariables)) and (Index<=High(FDateVariables)) then
  FDateVariables[Index]:=value
end;

function TXitamiResponse.GetIntegerVariable;
begin
 if (Index>=Low(FIntegerVariables)) and (Index<=High(FIntegerVariables)) then
  Result:=FIntegerVariables[Index]
 else
  Result:=0
end;

procedure TXitamiResponse.SetIntegerVariable;
begin
 if (Index>=Low(FIntegerVariables)) and (Index<=High(FIntegerVariables)) then
  FIntegerVariables[Index]:=value
end;

function TXitamiResponse.GetContent;
begin
 Result:=FContent
end;

procedure TXitamiResponse.SetContent;
begin
 FContent:=Value;
 ContentLength:=Length(FContent)
end;

function TXitamiResponse.GetStatusCode;
begin
 Result:=FStatusCode
end;

procedure TXitamiResponse.SetStatusCode;
begin
  if FStatusCode <> Value then
  begin
    FStatusCode := Value;
    ReasonString := StatusString(Value);
  end;
end;

function TXitamiResponse.GetLogMessage;
begin
 Result:=''
end;

procedure TXitamiResponse.SetLogMessage;
begin
end;

procedure TXitamiResponse.SendResponse;
var
  StatusString: string;
  Headers: string;
  I: Integer;

  procedure AddHeaderItem(const Item, FormatStr: string);
  begin
    if Item <> '' then
      Headers := Headers + Format(FormatStr, [Item]);
  end;

begin
  if HTTPRequest.ProtocolVersion <> '' then
  begin
    if (ReasonString <> '') and (StatusCode > 0) then
      StatusString := Format('%d %s', [StatusCode, ReasonString])
    else StatusString := '200 OK';
    AddHeaderItem(StatusString, 'Status: %s'#13#10);
    AddHeaderItem(Allow, 'Allow: %s'#13#10);
    for I := 0 to Cookies.Count - 1 do
      AddHeaderItem(Cookies[I].HeaderValue, 'Set-Cookie: %s'#13#10);
    AddHeaderItem(DerivedFrom, 'Derived-From: %s'#13#10);
    if Expires > 0 then
      Headers := Headers +
        FormatDateTime('"Expires: "' + DateFormat + ' "GMT"'#13#10, Expires);
    if LastModified > 0 then
      Headers := Headers +
        FormatDateTime('"Last-Modified: "' + DateFormat + ' "GMT"'#13#10, LastModified);
    AddHeaderItem(Title, 'Title: %s'#13#10);
    AddHeaderItem(WWWAuthenticate, 'WWW-Authenticate: %s'#13#10);
    AddCustomHeaders(Headers);
    AddHeaderItem(ContentVersion, 'Content-Version: %s'#13#10);
    AddHeaderItem(ContentEncoding, 'Content-Encoding: %s'#13#10);
    AddHeaderItem(ContentType, 'Content-Type: %s'#13#10);
    if (Content <> '') or (ContentStream <> nil) then
      AddHeaderItem(IntToStr(ContentLength), 'Content-Length: %s'#13#10);
    Headers := Headers + 'Content:'#13#10#13#10;
    HTTPRequest.WriteString(Headers);
  end;
  if ContentStream = nil then
    HTTPRequest.WriteString(Content)
  else if ContentStream <> nil then
  begin
    SendStream(ContentStream);
    ContentStream := nil; // Drop the stream
  end;
  FSent := True;
end;

procedure TXitamiResponse.SendRedirect(const URI: string);
begin
  HTTPRequest.WriteString(Format('Location: %s', [URI]));
  FSent := True;
end;

procedure TXitamiResponse.SendStream(AStream: TStream);
var
  Buffer: array[0..8191] of Byte;
  BytesToSend: Integer;
begin
  while AStream.Position < AStream.Size do
  begin
    BytesToSend := AStream.Read(Buffer, SizeOf(Buffer));
    FHTTPRequest.WriteClient(Buffer, BytesToSend);
  end;
end;

function TXitamiResponse.Sent;
begin
 Result:=FSent
end;

//-----------------------------------------------------------------------------
procedure TXitamiApplication.ConnectServer;
var i:integer;
begin
 i:=Pos(':',FServer);
 port.Connect(Title,Copy(FServer,1,i-1),Copy(FServer,i+1,maxint),Fhost);
end;

procedure TXitamiApplication.ProcessMessages;
var request:TWebRequest;
var response:TWebResponse;
var ok:boolean;
begin
 while port.AcceptRequest do begin
  request:=TXitamiRequest.Create(port.environment,port.post,port.data);
  try
   response:=TXitamiResponse.Create(request);
   try
    try
     ok:=HandleRequest(request,response);
     if ok then port.FinishRequest
    except
     // TODO:
     ok:=false
    end;
    if not ok then begin
     port.Disconnect;
     ConnectServer
    end
   finally
    response.Free
   end
  finally
   request.Free
  end
 end;
end;

type
 TWorker=class(TThread)
 public
  procedure Execute;override;
 end;

 procedure TWorker.Execute;
 begin
  TXitamiApplication(Application).ProcessMessages
 end;

procedure TXitamiApplication.OnWorkerSignal;
begin
 worker:=nil;
 PostQuitMessage(0)
end;

function GuardWndProc(Window: HWND; Message, WParam: Longint;
 LParam: Longint): Longint; stdcall;
begin
 case Message of
 WM_DESTROY:
  begin
   Result:=0;
   PostQuitMessage(0);
   exit
  end;
 end;
 Result:=DefWindowProc(Window,Message,wParam,lParam)
end;

procedure TXitamiApplication.Run;
var wnd:HWND;

 procedure PrepareWindow;
 const guardWindowClass='TGuardWindow';
 var cls:TWndClass;
 begin
  FillChar(cls,sizeof(cls),0);
  cls.hInstance:=hInstance;
  cls.lpszClassName:=guardWindowClass;
  cls.lpfnWndProc:=@GuardWndProc;
  RegisterClass(cls);
  wnd:=CreateWindowEx(WS_EX_APPWINDOW,guardWindowClass,nil,
     WS_POPUP or WS_CAPTION or WS_CLIPSIBLINGS or WS_SYSMENU or WS_MINIMIZEBOX,
     0,0,cw_usedefault,cw_usedefault,0,0,hInstance,nil);
 end;

 function GetParam(name:string; defaultValue:string):string;
 var i:integer;
 begin
  for i:=1 to ParamCount do begin
   Result:=ParamStr(i);
   if CompareText(Copy(Result,1,length(name))+'=',name+'=')=0 then begin
    Result:=Copy(Result,1+length(name)+1,maxint);
    exit
   end
  end;
  Result:=defaultValue
 end;

var msg:TMsg;
begin
 inherited;
 port:=TLRWP.Create(Self);
 if Title='' then Title:=GetParam('url',ChangeFileExt(ExtractFileName(ParamStr(0)),''));
 if FServer='' then FServer:=GetParam('server','127.0.0.1:81');
 if Fhost='' then Fhost:=GetParam('host','');
 ConnectServer;

 worker:=TWorker.Create(true);
 worker.FreeOnTerminate:=true;
 worker.OnTerminate:=OnWorkerSignal;
 worker.Resume;

 PrepareWindow;
 while GetMessage(msg,0,0,0) do begin
  if msg.message=WM_CLOSE then PostQuitMessage(0);
  TranslateMessage(msg);
  DispatchMessage(msg)
 end;

 if worker<>nil then begin
  port.Disconnect;
  worker.Suspend;
  worker.Free;
 end;
 worker:=nil
end;

procedure TXitamiApplication.Terminate;
begin
 if port<>nil then begin
  port.Disconnect
 end;
 if worker<>nil then begin
  worker.Suspend;
  worker.Free;
  worker:=nil
 end;
end;

procedure InitApplication;
begin
  Application := TXitamiApplication.Create(nil);
end;

procedure DoneApplication;
begin
//  if Assigned(Application) then TXitamiApplication(Application).guard.Free;
  Application.Free;
  Application := nil;
end;

initialization
  InitApplication;
finalization
  DoneApplication;
end.
