{ VERSION: 1.3 }
{ 1.3 / 22.3.2005
  - nova funkce ValidFile(): Boolean
  1.2 / 3.2.2005
  - nova vlastnost TPackFileSys.Password
    (aby to fungovalo musi ji pouzivat Compress() a Decompress())
}
unit PackFileSys;

interface
uses Classes,SysUtils,fileSys;

 const
  PACK_STATUS_OK              =  0;
  PACK_STATUS_UNKNOWN_ERROR   = -1;
  PACK_STATUS_CRC_OR_PASSWORD = -2;
 type
  EBadPassword=class(Exception) end;
  TPackFileSystem=class(TFileSystem)
  private
   Lock:TObject;
  protected
   TempHome:string;
   ArchiveName:string;
   OpenMode:Word;
  private
   FPassword: string;            // current password
   FOpenPassword: string;        // password used to open archive
   FPasswordChanged: Boolean;    // password used to open archive differs from new password
   procedure SetPassword(const Value: string);
  protected
   property PasswordChanged: Boolean read FPasswordChanged;
  protected
   function Decompress(const ArchiveName:string; const TempHome:string):integer;virtual;
   function Compress(const ArchiveName:string; const TempHome:string; update:boolean):integer;virtual;
  protected
   FCurrentDir:string;
   function DoSetCurrentDir(const Dir:string):boolean;override;
   function MapFileName(const fileName:string):string;
   destructor Destroy;override;
  public
   constructor Create(const AnArchiveFileName:string; Mode:Word; const APassword: string = '');
   function  GetCurrentDir:string;override;
   function  DirectoryExists(const Dir: string): Boolean;override;
   function  CreateFileMapping(const FileName:string; targetFileSystem:TFileSystem):TFileMapping;override;
   function  GetFileStream(const FileName: string; Mode: Integer): TStream;override;
   function  FindFirst(const Path: string; Attr: Integer; var F: TSearchRec): Integer;override;
   function  FindNext(var F: TSearchRec): Integer;override;
   procedure FindClose(var F: TSearchRec);override;
   function DeleteFile(const FileName: string): Boolean;override;
   function FileSetAttr(const FileName: string; Attr: Integer): Integer;override;
   function FileSetDate(const FileName:string; Age:integer):Integer;override;
   function CreateDir(const Dir: string): Boolean;override;
   function RemoveDir(const Dir: string): Boolean;override;
   class function ValidFile(const AnArchiveFileName:string): Boolean; virtual;
  public
   property Password: string read FPassword write SetPassword;
  end;

implementation
uses fileUtils;

const
 RootDrive='C:\'; { TUNE: }

 procedure RaiseError(ErrorCode:integer);
 var E:EInOutError;
 begin
  RaiseLastWin32Error;
  { if we get here, no above didn't raise }
  E:=EInOutError.Create('');
  E.ErrorCode:=ErrorCode;
  raise E
 end;

 procedure Check(condition:boolean; ErrorCode:integer);
 begin
  if not condition then RaiseError(ErrorCode)
 end;

 procedure CheckStatus(status: Integer);
 begin
   case status of
   PACK_STATUS_OK:
     ;
   PACK_STATUS_CRC_OR_PASSWORD:
     raise EBadPassword.Create('Incorrect password' { TUNE: });
     
   else
     Check(False, status)
   end
 end;

type
 TLock=class
 private
  Mode:Word;
  FileName:string;
  Stream:TStream;
 public
  constructor Create(const AFileName:string; AMode:Word);
  destructor Destroy;override;
 end;

 constructor TLock.Create;
 begin
  inherited Create;
  FileName:=AFileName;
  Mode:=AMode;
  Stream:=TFileStream.Create(FileName,Mode);
 end;

 destructor TLock.Destroy;
 begin
  Stream.Free;
  if Mode=fmCreate then
   DeleteFile(FileName);
  inherited
 end;

 constructor TPackFileSystem.Create;
 var status:integer;
 begin
  inherited Create;
  FPassword := APassword;
  FOpenPassword := APassword;

  FCurrentDir:=RootDrive;
  ArchiveName:=AnArchiveFileName;
  OpenMode:=Mode;

  TempHome:=OSFileSystem.GetTempFileName('\');
  Check(TempHome<>'',3);
  if (Mode and fmCreate)=fmCreate then begin
   OpenMode:=fmOpenWrite;
  end else begin
   OpenMode:=Mode and $F;
  end;

  try
   case OpenMode of
   fmOpenRead:
    ;
   fmOpenReadWrite,
   fmOpenWrite:
    begin
     // first check if archive is writable
     if OSFileSystem.FileExists(ArchiveName) then
      TLock.Create(ArchiveName, fmOpenReadWrite).Free;

     // gain exclusive access
     Lock:=TLock.Create(ArchiveName+'.lck',fmCreate);

     
    end;
   end;

   // get data
   case OpenMode of
   fmOpenRead,
   fmOpenReadWrite:
    begin
     if OSFileSystem.FileExists(ArchiveName) then begin
      status:=Decompress(ArchiveName,TempHome);
      CheckStatus(status)
     end else
      // use blank
      ;
    end;
   end
  except
   OpenMode:=fmOpenRead;      // disable write-back
   raise
  end
 end;

 destructor TPackFileSystem.Destroy;

  procedure Pack(update:boolean);
  var status:integer;
  begin
   status:=Compress(ArchiveName,TempHome,update);
   CheckStatus(status)
  end;

 begin
  inherited;
  Lock.Free;
  Lock:=nil;
  if TempHome<>'' then begin
   try
    case OpenMode of
    fmOpenReadWrite:
     Pack(true);
    fmOpenWrite:
     Pack(false);
    end
   finally
    EraseFolder(OSFileSystem,TempHome)
   end
  end
 end;

 class function TPackFileSystem.ValidFile(const AnArchiveFileName:string): Boolean;
 begin
  Result := False
 end;

 function TPackFileSystem.DoSetCurrentDir;
 begin
  FCurrentDir:=Dir;
  Result:=true
 end;

 function TPackFileSystem.GetCurrentDir;
 begin
  Result:=FCurrentDir
 end;

 function TPackFileSystem.DirectoryExists;
 begin
  if CompareText(Dir,RootDrive)=0 then
   Result:=true
  else
   Result:=inherited DirectoryExists(Dir)
 end;

 function TPackFileSystem.MapFileName;
 begin
  Result:=TempHome+'\'+Copy(ExpandFileName(fileName),length(RootDrive)+1,maxint)
 end;

 procedure TPackFileSystem.FindClose;
 begin
  OSFileSystem.FindClose(F)
 end;

 function TPackFileSystem.FindFirst;
 begin
  Result:=OSFileSystem.FindFirst(MapFileName(Path),Attr,F)
 end;

 function TPackFileSystem.FindNext;
 begin
  Result:=OSFileSystem.FindNext(F)
 end;

 function TPackFileSystem.GetFileStream;
 begin
  Result:=OSFileSystem.GetFileStream(MapFileName(fileName),Mode)
 end;

 function TPackFileSystem.CreateFileMapping;
 var fullName:string;
 begin
  if targetFileSystem=OSFileSystem then begin
   try
    fullName:=MapFileName(fileName);
    if not OSFileSystem.FileExists(fullName) then
     OSFileSystem.GetFileStream(fullName,fmCreate).Free
   except
    Result:=nil;
    exit
   end;
   Result:=TFileMapping.Create;
   Result.source.fileSystem:=Self;
   Result.source.fileName:=ExpandFileName(fileName);
   Result.target.fileSystem:=OSFileSystem;
   Result.target.fileName:=fullName
  end else
   Result:=inherited CreateFileMapping(fileName,targetFileSystem)
 end;

 function TPackFileSystem.DeleteFile;
 begin
  Result:=OSFileSystem.DeleteFile(MapFileName(fileName))
 end;

 function TPackFileSystem.FileSetAttr;
 begin
  Result:=OSFileSystem.FileSetAttr(MapFileName(fileName),Attr)
 end;

 function TPackFileSystem.FileSetDate;
 begin
  Result:=OSFileSystem.FileSetDate(MapFileName(fileName),Age)
 end;

 function TPackFileSystem.CreateDir;
 begin
  Result:=OSFileSystem.CreateDir(MapFileName(Dir))
 end;

 function TPackFileSystem.RemoveDir;
 begin
  Result:=OSFileSystem.RemoveDir(MapFileName(Dir))
 end;

 function TPackFileSystem.Decompress;
 begin
  Result:=PACK_STATUS_UNKNOWN_ERROR
 end;

 function TPackFileSystem.Compress;
 begin
  Result:=PACK_STATUS_UNKNOWN_ERROR
 end;

 procedure TPackFileSystem.SetPassword(const Value: string);
 begin
  FPassword := Value;
  FPasswordChanged := FPassword <> FOpenPassword
 end;

end.
