{ VERSION: 1.5 }
{ 1.5 / 4.6.2006
  - ere i soubor nulov dlky ( viz. Compress() )
  1.4 / 22.3.2005
  - implementuje ValidFile()
  1.3 / 3.2.2005
  - podporuje sifrovani TPackFileSys.Password
  1.2 / 6.12.2004
  - datum souboru je podle nejnovjho souboru vevnit
  1.1 / 3.12.2004
  - varovani se vraci jako OK
}
unit RARFileSys;

interface
uses Classes,SysUtils,fileSys,PackFileSys;

 type
  TCompressLevel=(clDefault,clFast,clMedium,clBest);
  TRARFileSystem=class(TPackFileSystem)
  public
   CompressLevel:TCompressLevel;
  private
   Compressor:string;
   function RAR(const ArchiveName:string; const TempHome:string; const commands:string):integer;
  protected
   function Decompress(const ArchiveName:string; const TempHome:string):integer;override;
   function Compress(const ArchiveName:string; const TempHome:string; update:boolean):integer;override;
   destructor Destroy;override;
  public
   class function ValidFile(const AnArchiveFileName:string): Boolean; override;
  end;

implementation
uses Windows,fileUtils;

 function ExecEx(command:string; wait:boolean; ShowCmd:integer; currentDir:string):integer;
 var
   zFileName: array[0..MAX_PATH*2] of Char;
   zCurrentDir: array [0..MAX_PATH] of Char;
   pCurrentDir: PChar;

   startup:TStartupInfo;
   process:TProcessInformation;

   ok:boolean;
   ErrNo:integer;
 begin
  if currentDir='' then
   pCurrentDir:=nil
  else
   pCurrentDir:=StrPCopy(zCurrentDir,currentDir);

  FillChar(startup,sizeof(startup),0);

  with startup do begin
   cb:=sizeof(startup);
   dwFlags:=STARTF_USESHOWWINDOW;
   wShowWindow:=ShowCmd
  end;

  ok:=CreateProcess(
   nil,                            // aplikace
   StrPCopy(zFileName,command),  // prikazovy radek
   nil,
   nil,
   false,
   (CREATE_NEW_PROCESS_GROUP or NORMAL_PRIORITY_CLASS or DETACHED_PROCESS),
   nil,
   pCurrentDir,
   startup,
   process);

  if not ok then begin
   ErrNo:=GetLastError;
   raise Exception.CreateFmt('Exec("%s")=%d'#13#13'%s',[zFileName,ErrNo,SysErrorMessage(ErrNo)]);
  end;

  CloseHandle(process.hThread);
  Result:=process.hProcess;
 end;

 class function TRARFileSystem.ValidFile(const AnArchiveFileName:string): Boolean;
 var
   Stream: TStream;
   Signature: string;
 begin
   Result := False;
   try
     Stream := OSFileSystem.GetFileStream(AnArchiveFileName, fmOpenRead or fmShareDenyWrite);
     try
       if Stream.Size >= 7 then begin
         SetLength(Signature, 7);
         Stream.ReadBuffer(Signature[1], 7);
         Result := Signature = 'Rar!'#$1A#$07#$00
       end
     finally
       Stream.Free
     end
   except
   end
 end;

{$R RARFileSys.res}

 function TRARFileSystem.RAR;
 var handle:THandle;
 var status:{$IFDEF VER150}Cardinal{$ELSE}integer{$ENDIF};
 var stream:TResourceStream;
 begin
  if Compressor='' then begin
   Compressor:=OSFileSystem.GetTempFileName('exe');
   try
    stream:=TResourceStream.Create(hInstance,'RAREXE',RT_RCDATA);
    try
     stream.SaveToFile(Compressor);
    finally
     stream.Free
    end
   except
    Compressor:='';
    raise
   end
  end;

  handle:=ExecEx(Format('"%s" %s "%s" *.*',[Compressor,commands,ArchiveName]),false,sw_hide,CreateFileName(nil,TempHome,'',false));
  WaitForSingleObject(handle,INFINITE);
  status:=THandle(-1);
  GetExitCodeProcess(handle,status);
  CloseHandle(handle);
  // status 0 = ok, 1 = warning (no errors)
  if status in [0,1] then
    Result := PACK_STATUS_OK
  else
  if status in [3] then
    Result := PACK_STATUS_CRC_OR_PASSWORD
  else
    Result:=status
 end;

 function TRARFileSystem.Decompress;
 var
  opt: string;
 begin
  if Password = '' then
   opt := '-p-'
  else
   opt := '"-p'+Password+'"';
  Result:=RAR(ArchiveName,TempHome,'x '+opt)
 end;

 function TRARFileSystem.Compress;
 var opt:string;
 var bopt:string;
 begin
  if PasswordChanged then
   // we must repack
   update := False;
  if FileSize(ArchiveName) = 0 then
   // invalid file format ( just a formal file )
   update := False;

  case CompressLevel of
  clFast:
   opt:='-m0';
  clMedium:
   opt:='-m3';
  clBest:
   opt:='-m5';
  else
   opt:=''
  end;
  if Password <> '' then
   opt:=opt + ' "-hp'+Password+'"';
  bopt:=opt;
  case CompressLevel of
  clBest:
   bopt:=bopt+' -s';
  end;
  if update then
   Result:=RAR(ArchiveName,TempHome,'a -u -r -md4096 -tl '+opt)
  else begin
   OSFileSystem.DeleteFile(ArchiveName);
   Result:=RAR(ArchiveName,TempHome,'a -r -md4096 -tl '+bopt)
  end
 end;

 destructor TRARFileSystem.Destroy;
 begin
  try
   inherited Destroy;
  finally
   if Compressor<>'' then
    OSFileSystem.DeleteFile(Compressor)
  end
 end;

end.
