{ VERSION: 1.0 }
{ 1.0 / 8.2.2005
  - funkce Timer..() pro mereni s milisekundovou presnosti. funkce jsou
    odolne na preteceni systemoveho pocitadla GetTickCount()
}

unit SimpleTimers;

interface

type
  PTimer = ^TTimer;
  TTimer = record
    Enabled: Boolean;
    Ticks: Int64;
    LastTickCount: Cardinal;
  end;

// pripravi Timer pro pouziti
procedure TimerReset(var Timer: TTimer);

// nastartuje odpocitavani ( pripadne od urciteho okamziku z minulosti )
procedure TimerStart(var Timer: TTimer); overload;
procedure TimerStart(var Timer: TTimer; StartTickCount: Cardinal); overload;

// vynuluje pamet a (znovu) nastartuje odpocitavani ( pripadne od urciteho okamziku z minulosti )
procedure TimerReStart(var Timer: TTimer); overload;
procedure TimerReStart(var Timer: TTimer; StartTickCount: Cardinal); overload;

// zastavi odpocitavani
procedure TimerStop(var Timer: TTimer); overload;
procedure TimerStop(var Timer: TTimer; StopTickCount: Cardinal); overload;

// vrati pocet odmerenych milisekund
function  TimerValue(var Timer: TTimer): Cardinal;

implementation
uses
  Windows;

procedure TimerReset(var Timer: TTimer);
begin
  Timer.Enabled := False;
  Timer.Ticks := 0;
  Timer.LastTickCount := 0
end;

procedure TimerStart(var Timer: TTimer);
begin
  // TODO: co kdyz mezi stop a start ubehne nejaky tick ?
  TimerStop(Timer);
  Timer.LastTickCount := GetTickCount;
  Timer.Enabled := True
end;

procedure TimerStart(var Timer: TTimer; StartTickCount: Cardinal);
begin
  // TODO: co kdyz mezi stop a start ubehne nejaky tick ?
  TimerStop(Timer);
  Timer.LastTickCount := StartTickCount;
  Timer.Enabled := True
end;

procedure TimerReStart(var Timer: TTimer);
begin
  Timer.Ticks := 0;
  Timer.LastTickCount := GetTickCount;
  Timer.Enabled := True
end;

procedure TimerReStart(var Timer: TTimer; StartTickCount: Cardinal);
begin
  Timer.Ticks := 0;
  Timer.LastTickCount := StartTickCount;
  Timer.Enabled := True
end;

procedure TimerStop(var Timer: TTimer);
begin
  if Timer.Enabled then
    TimerStop(Timer, GetTickCount)
end;

procedure TimerStop(var Timer: TTimer; StopTickCount: Cardinal);
begin
  if Timer.Enabled then begin
    Timer.Enabled := False;

    if StopTickCount >= Timer.LastTickCount then
      Timer.Ticks := Timer.Ticks + (StopTickCount - Timer.LastTickCount)
    else begin
      // doslo k preteceni GetTickCount pocitadla
      Timer.Ticks := Timer.Ticks + (High(Cardinal) - Timer.LastTickCount) + 1 + StopTickCount
    end
  end
end;

function  TimerValue(var Timer: TTimer): Cardinal;
begin
  if Timer.Enabled then
    // prenastavime LastTickCount jako prevenci pred pretecenim GetTickCount a
    // zaktualizuje Timer.Ticks
    TimerStart(Timer);

  Result := Timer.Ticks;
  if Result <> Timer.Ticks then
    // doslo k preteceni pocitadla - vratime maximalni vratitelnou hodnotu
    Result := High(Cardinal)
end;

end.
