Unit LZRW1;

{$A+} { word align }
{$O+} { ?? ?? }
{$Q-}

{*****************************************************************************
*
*TLZRW1 file compression component.
*----------------------------------
* disconnected the visual feedback and streams as the only option - 20000912 - m.n.
*
*Compresses a file with :
*------------------------
*
*    either the LZRW1/KH or LZH compression algorithm,
*           with code posted by Kurt Haenen on the SWAG (lzrw1).
*    or the Japanese LZH compression algorithm
*           ( LZSS coded by Haruhiko OKUMURA
*             Adaptive Huffman Coding coded by Haruyasu YOSHIZAKI
*             Edited and translated to English by Kenji RIKITAKE
*             Translated from C to Turbo Pascal by Douglas Webb   2/18/91
*             posted by Doug Webb on the SWAG (preskit2\lzh).)
*
*
*Visual feedback on a Panel if so desired.
*
*All VCL code by D. Heijl , may 8-9 1996
*
*The Getblock/PutBlock procedures are based on the code in
*lzhtest.pas by Douglas Webb.
*
*
*The files lzh.pas and lzrw1kh.pas are essentially untouched
*(only some cosmetic changes, also added exceptions)
*
*--------------------------------------------------------------------
* V2.00.00 :
*
* Code for using a Stream  instead of a File added by Stefan Westner
*                          25 May 1997 (stefan.westner@stud.uni-bamberg.de)
* I removed the seeks to the beginning of the stream (except for the auto guess)
* and the Steeam.Clear call, so that you have more freedom using TFileStream.
*                          30 May 1997 (Danny.Heijl@cevi.be)
* I also added a "Threshold" property that dictates the behaviour of Advise.
*--------------------------------------------------------------------
*
* Feel free to use or give away this software as you see fit.
* Just leave the credits in place if you alter the source.
*
* This software is delivered to you "as is",
* no guarantees, it may blow up or trigger World War Three
* for all I know.
*
* If you find any bugs and let me know, I will try to fix them.
*
* I believe in programmers around the world helping each other
* out by distributing free source code.
*
*Danny Heijl, may 10 1996.
*Danny.Heijl@cevi.be
*
*----------------------------------------------------------------
*****************************************************************}

interface

uses SysUtils, Classes, Lzrw1kh, Lzh;

{$IFDEF WIN32}
type
    Int16   = SmallInt;
    SString = ShortString;
{$ELSE}
type
    Int16   = Integer;
{$ENDIF}

CONST
  LZRWIdentifier : LONGINT =
  ((((((ORD('L') SHL 8)+ORD('Z')) SHL 8)+ORD('R')) SHL 8)+ORD('W'));

  LZHIdentifier : LONGINT =
  ((((((ORD('L') SHL 8)+ORD('Z')) SHL 8)+ORD('H')) SHL 8)+ORD('!'));

  ChunkSize = Lzrw1kh.BufferMaxSize;
  IOBufSize = (ChunkSize + 16);


type
     ELzrw1Exception = class(Exception);
     TCompressMode = (Good, Fast, Auto);
     LZHBuf  = Array[1..ChunkSize] OF BYTE;
     PLZHBuf = ^LZHBuf;

type
  Tlzrw1 = class(TComponent)
  private
    FLZH                         : TLZH;
    FLZR                         : TLZR;
    FCompressMode                : TCompressMode;

    FInputStream                 : TStream;
    FOutputStream                : TStream;

    FThreshold                   : Integer;  { autoguess "fast" threshold }

    LZHInBuf, LZHOutBuf : PLZHBuf;
    SRCBuf,DSTBuf       : BufferPtr;    { defined in LZRW1KH }

    SRCSize,DSTSize     : LongInt;

    Tmp                 : Longint;
    Identifier          : LONGINT;
    CompIdentifier      : LONGINT;
    InSize,OutSize      : LONGINT;

    Size : Longint;

    Buf : Longint; { getblock }
    PosnR : Word;  { getblock }
    PosnW : Word;  { putblock }

    procedure CheckWrite(Actual, Desired : Longint);
    procedure CheckRead(Actual,  Desired : Longint);

    procedure GetBlock(VAR Target; NoBytes:Word; VAR Actual_Bytes:Word);
    procedure PutBlock(VAR Source; NoBytes:Word; VAR Actual_Bytes:Word);

    procedure LZrw1Compress;
    procedure LZrw1Decompress;
    procedure LZHCompress;
    procedure LZHDecompress;

  protected
    Function CompressFile   : Longint;
    Function DeCompressFile : Longint;
    function GetBestMode    : TcompressMode;

  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    property InputStream  : TStream read FInputStream   write FInputStream;
    property OutputStream : TStream read FOutputStream  write FOutputStream;

  published
    Function Compress     : LongInt;
    Function Decompress   : Longint;
    Function Advise       : TcompressMode;
    property CompressMode : TCompressMode read  FCompressMode
                                          write FcompressMode default Good;
    property Threshold    : Integer read  FThreshold
                                    write Fthreshold  default 40;
end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Compon',[Tlzrw1]);
end;

constructor TLzrw1.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
                            { initialize defaults }
  FcompressMode     := Good;
  FThreshold        := 40;      { 40 % autoguess default threshold }
  FLZH              := TLZH.Create;
  FLZR              := TLZR.Create;
end;

Destructor TLzrw1.Destroy;
Begin
  FLZH.Destroy;
  FLZR.Destroy;
  inherited Destroy;
end;

           { the 2 execute methods : compress and decompress }
           {-------------------------------------------------}

function TLzrw1.Compress : Longint;
begin
  Result := CompressFile;                   { compress stream/file }
end;


function TLzrw1.DeCompress : Longint;
begin
  Result := DeCompressFile;                 { decompress stream/file }
end;

          { the 3d execute method : advise compression method }
          {---------------------------------------------------}

function TLzrw1.Advise : TcompressMode;
begin
  Result := GetBestMode;                    { get a guess on file/stream }
end;

          { some common subroutine functions }
          {----------------------------------}
const
  SLZRW1FILERIO   = 'Lzrw1 : Error reading from input file : ';
  SLZRW1STREAMRIO = 'Lzrw1 : Error reading from input stream !';
  SLZRW1FILEWIO   = 'Lzrw1 : Error writing to output file : ';
  SLZRW1STREAMWIO = 'Lzrw1 : Error writing to output stream !';

{ Check if Write was successfull, raise an exception if not }
procedure TLzrw1.CheckWrite(Actual, Desired : Longint);
begin
  if (Actual <> Desired) then begin
    Raise ELzrw1Exception.Create(SLZRW1STREAMWIO)
  end;
end;

{ check if Read was successfull, raise an exception if not }
procedure TLzrw1.CheckRead(Actual, Desired : Longint);
begin
  if (Actual <> Desired) then begin
    Raise ELzrw1Exception.Create(SLZRW1STREAMRIO)
  end;
end;

          { the LZH reader and writer procedures }
          {--------------------------------------}

const
  SLZHIN   = '(LZH) Bytes in : ';
  SLZHOUT  = ' Bytes out : ';

          { the reader : GetBlock }

Procedure TLzrw1.GetBlock(VAR Target; NoBytes:Word; VAR Actual_Bytes:Word);
begin
  if (PosnR > Buf) or (PosnR + NoBytes > SUCC(Buf)) then begin
    if PosnR > Buf then begin
       Buf := FInputStream.Read(LZHInBuf^,ChunkSize);
       if (Buf < 0) then begin
         Raise ELzrw1Exception.Create(SLZRW1STREAMRIO + ' (LZH)')
       end;
       INC(InSize,Buf);
    end else begin
      Move(LZHInBuf^[PosnR],LZHInBuf^[1],Buf - PosnR);
      Tmp := FInputStream.Read(LZHInBuf^[Buf-PosnR],ChunkSize - (Buf - PosnR));
      if (Tmp < 0) then begin
        Raise ELzrw1Exception.Create(SLZRW1STREAMRIO + ' (LZH)')
      end;
      INC(InSize,Tmp);
      Buf := Buf - PosnR + Tmp;
    end;

    if Buf = 0 then begin
       Actual_Bytes := 0;
       Exit;
    end;

    PosnR := 1;
  end;

  Move(LZHInBuf^[PosnR],Target,NoBytes);
  INC(PosnR,NoBytes);

  if PosnR > SUCC(Buf) then
    Actual_Bytes := NoBytes - (PosnR - SUCC(Buf))
  else
    Actual_Bytes := NoBytes;

end;

          { and the writer : PutBlock }

Procedure TLzrw1.PutBlock(VAR Source; NoBytes:Word; VAR Actual_Bytes:Word);

begin
  If NoBytes = 0 then begin   { Flush condition }
    Tmp := FOutputStream.Write(LZHOutBuf^,PRED(PosnW));
    CheckWrite(Tmp, PRED(PosnW));
    Inc(OutSize, Tmp);
    EXIT;
  end;
  if (PosnW > ChunkSize) or (PosnW + NoBytes > SUCC(ChunkSize)) then begin
    Tmp := FOutputStream.Write(LZHOutBuf^,PRED(PosnW));
    CheckWrite(Tmp, PRED(PosnW));
    Inc(OutSize, Tmp);
    PosnW := 1;
  end;
  Move(Source,LZHOUTBuf^[PosnW],NoBytes);
  INC(PosnW,NoBytes);
  Actual_Bytes := NoBytes;
end;

                { compress a file with LZRW1/KH (FAST) }
                {--------------------------------------}

const
  SLZRW1KHIN  = '(LZRW1/KH) Bytes in : ';
  SLZRW1KHOUT = ' Bytes out : ';
  SLZRW1KHSIZEEXC =
     'Lzrw1 (decompression LZRW1/KH) : Original and compressed sizes do not match !';

Procedure TLzrw1.LZRW1Compress;
begin                                           { start compressing }
  SRCSize := ChunkSize;
  InSize := 0;
  WHILE (SRCSize = ChunkSize) DO begin
                                      { read a block af data }
    SrcSize := FInputStream.Read(SrcBuf^, ChunkSize);

                                  { this fix is BAD, and not needed because    }
                                  { LZRW1KH.PAS handles the problem now        }
{    if (SrcSize = 0) then exit;} {      fix of bug in decompression discovered}
{                               } {      by <Domus@compuserve.com> }

    if (SrcSize < 0) then begin
      Raise ELzrw1Exception.Create(SLZRW1STREAMRIO + ' (LZRW1/KH)')
    end;
    INC(InSize,SRCSize);
                                      { compress it }
    DSTSize := FLZR.Compression(SRCBuf,DSTBuf,SRCSize);
                                      { write out compressed size }
    Tmp := FOutputStream.Write(DstSize, SizeOf(Word));
    CheckWrite(Tmp, Sizeof(Word));
    INC(OutSize,Tmp);
                                       { write out compressed data }
    Tmp := FOutputStream.Write(DstBuf^, DstSize);
    CheckWrite(Tmp, DstSize);
    INC(OutSize,Tmp);

  end;       { endwhile SRCSize = ChunkSize }
end;

                { compress a file with LZH (GOOD) }
                {---------------------------------}

procedure TLzrw1.LZHCompress;
var
  Bytes_Written : Longint;
  Temp          : Word;

begin

                    { initialize put/getblock variables }
  Buf := 0;
  PosnR := 1;
  PosnW := 1;
                    { pack the file with LZH }
  FLZH.LZHPack(Bytes_written, GetBlock, PutBlock);

                    { flush last buffer }
  PutBlock(Size, 0, Temp);

end;

                { decompress a file with LZRW1 (FAST) }
                {-------------------------------------}

procedure TLzrw1.LZRW1Decompress;
var
  OrigSize : Longint;

begin
                    { read in uncompressed filesize }
  Tmp := FInputStream.Read(OrigSize, sizeof(Longint));
  CheckRead(Tmp, Sizeof(Longint));
  Inc(InSize,SIZEOF(LONGINT));
                                         { start decompression }
  WHILE (DSTSize = ChunkSize) DO begin
                                         { read size of compressed block }
    Tmp := FInputStream.Read(SrcSize, SizeOf(Word));
    CheckRead(Tmp, Sizeof(Word));
                                          { read compressed block }
    Tmp := FInputStream.Read(SrcBuf^, SrcSize);
    Checkread(Tmp, SrcSize);
    INC(InSize,Tmp + SIZEOF(WORD));
                                           { decompress block }
    DSTSize := FLZR.Decompression(SRCBuf,DstBuf,SRCSize);
                                           { write it out }
    Tmp := FOutputStream.Write(DstBuf^, DstSize);
    CheckWrite(Tmp, DstSize);
    INC(OutSize,Tmp);
  end;   { endwhile data to read }

  If (OutSize <> OrigSize) then
    Raise ELzrw1Exception.Create(SLZRW1KHSIZEEXC);

end;

                { decompress a file with LZH (GOOD) }
                {-----------------------------------}

const
  SLZHSIZEEXC =
     'Lzrw1 (LZH decompression) : Original and compressed sizes do not match !';

procedure TLzrw1.LZHDecompress;
var
  OrigSize      : Longint;
  Temp          : Word;

begin

                    { read in uncompressed filesize }
  Tmp := FInputStream.Read(OrigSize, sizeof(Longint));

  CheckRead(Tmp, Sizeof(Longint));
  Inc(InSize,SIZEOF(LONGINT));

                    { initialize put/getblock variables }
  PosnR := 1;
  Buf   := 0;
  PosnW := 1;
                    { unpack the file with LZH }
  FLZH.LZHUnPack(OrigSize, GetBlock, PutBlock);

                    { flush last buffer }
  PutBlock(Size, 0, Temp);

  If (OutSize <> OrigSize) then begin
    Raise ELzrw1Exception.Create(SLZHSIZEEXC);
  end;

end;

                { the main code common to both (de)compression methods  }
                {-------------------------------------------------------}

const
  SNOMEMORY       = 'Lzrw1 (de)compression : Failed to get memory for I/O buffers !';
  SNOSIZESTREAM   = 'Lzrw1 compression : Failed to obtain the size of the input stream !';
  SNOSIZEFILE     = 'Lzrw1 compression : Failed to obtain the size of : ';
  SINSTREAMERR    = 'Lzrw1 (de)compression : Failed to initialize input stream !';
  SOUTSTREAMERR   = 'Lzrw1 (de)compression : Failed to initialize output stream !';
  SINFILEERR      = 'Lzrw1 (de)compression : Failed to open input file : ';
  SOUTFILEERR     = 'Lzrw1 (de)compression : Failed to open output file : ';
  SINSTREAMFMT    = 'Lzrw1 decompression : input stream contains no valid header !';
  SINFILEFMT      = ' : Lzrw1 decompression : this file has no valid header !';
  SINSTREAMSTATC  = 'Lzrw1 : inputstream successfully compressed at ';
  SINFILESTATC    = ' : Lzrw1 : successfully compressed at ';
  SINSTREAMSTATD = 'LZrw1 : inputstream successfully decompressed at ';
  SINFILESTATD   = ' : Lzrw1 : successfully decompressed at ';

                { compress a file }
                {-----------------}

function TLzrw1.CompressFile : Longint;
var
  Mode : TcompressMode;

begin

  If (FcompressMode = Auto) then
    Mode := GetBestMode
  else
    Mode := FcompressMode;

  try
    Getmem(SRCBuf, IOBufSize);
    Getmem(DSTBuf, IOBufSize);
    LZHInBuf := PLZHBuf(SRCBuf);
    LZHOutBuf := PLZHBuf(DSTBuf);
  except
    Raise ELzrw1Exception.Create(SNOMEMORY);
  end;


  if (Mode = Fast) then
    CompIdentifier := LZRWIdentifier
  else
    CompIdentifier := LZHIdentifier;

  try
    try
      Size:=FInputStream.Size
    except
      ELzrw1Exception.Create(SNOSIZESTREAM)
    end;

    try                    { try to open the streams/files }
      if FInputStream.Seek(0, soFromCurrent) < 0 then
        Raise ELzrw1Exception.Create(SINSTREAMERR);
      if FOutputStream.Seek(0, soFromCurrent) < 0 then
        Raise ELzrw1Exception.Create(SOUTSTREAMERR);

      try               { try to compress the file }
                        { write out compression ID }
        Tmp := FOutputStream.Write(CompIdentifier, sizeof(Longint));
        CheckWrite(Tmp, Sizeof(Longint));
        OutSize := SIZEOF(LONGINT);
        InSize := 0;
                          { write out uncompressed filesize }
        Tmp := FOutputStream.Write(Size, sizeof(Longint));
        CheckWrite(Tmp, Sizeof(Longint));
        Inc(OutSize,SIZEOF(LONGINT));

        if (Mode = Fast) then
          LZRW1Compress
        else
          LZHCompress;

      except                  { error while compressing }
                              { leave streams alone , delete outputfile }
        Raise;                { and reraise to inform user }
      end;

    finally
    end;

  finally
    Freemem(SRCBuf,IOBufSize);
    Freemem(DSTBuf,IOBufSize);
  end;

  Result := OutSize;

end;


                { decompress a file }
                {-------------------}

function TLzrw1.DeCompressFile : Longint;

begin

  try
    Getmem(SRCBuf, IOBufSize);
    Getmem(DSTBuf, IOBufSize);
    LZHInBuf := PLZHBuf(SRCBuf);
    LZHOutBuf := PLZHBuf(DSTBuf);
  except
    Raise ELzrw1Exception.Create(SNOMEMORY);
    exit;
  end;

  try
    try
      if FInputStream.Seek(0, soFromCurrent)<0 then
         Raise ELzrw1Exception.Create(SINSTREAMERR);
      if FOutputStream.Seek(0, soFromCurrent)<0 then
         Raise ELzrw1Exception.Create(SOUTSTREAMERR);

      try
                                          { read compression ID }
        Tmp := FInputStream.Read(Identifier, Sizeof(Longint));
        CheckRead(Tmp, Sizeof(Longint));
        if (Identifier <> LZRWIdentifier)
             and (Identifier <> LZHIdentifier) then begin
          Raise ELzrw1Exception.Create(SINSTREAMFMT)
        end;
        DSTSize := ChunkSize;
        InSize := SIZEOF(LONGINT);
        OutSize := 0;

        if (Identifier = LZRWIdentifier) then
          LZRW1Decompress
        else
          LZHDecompress;

      except
        Raise
      end;

    finally
    end;

  finally
    Freemem(SRCBuf,IOBufSize);
    Freemem(DSTBuf,IOBufSize);
  end;

  Result := OutSize;

end;

                { Guess the best compression mode }
                { returns Good or Fast }
const
  SGUESSSTRM = 'Guessing the best compression mode for the inputstream';
  SGUESSFILE = 'Guessing the best compression mode for : ';
  SGUESSMEMERR      = 'Lzrw1 advise : Error getting I/O buffers';
  SGUESSSTRMSIZEERR = 'Lzrw1 advise : can not get size of inputstream !';
  SGUESSFILESIZEERR = 'Lzrw1 advise : can not get the size of : ';
  SGUESSSTRMOPENERR = 'Lzrw1 edvise : can not initialize the input stream!';
  SGUESSFILEOPENERR = 'Lzrw1 advise : can open input file : ';
  SGUESSSTRMRDERR = 'Lzrw1 advise : input stream read error !';
  SGUESSFILERDERR = 'Lzrw1 advise : read error on input file : ';
  SGUESSFAST = 'Lzrw1 recommends FAST compression : ';
  SGUESSGOOD = 'Lzrw1 recommends GOOD compression : ';

function TLzrw1.GetBestMode: TCompressMode;
var
  CompressedSize : Longint;
  UncompressedSize : Longint;
  SavedStreamPos : Longint;

  begin

  Result := Good;

  SavedStreamPos := FInputStream.Position;

  try
    Getmem(SRCBuf, IOBufSize);
    Getmem(DSTBuf, IOBufSize);
  except
    Raise ELzrw1Exception.Create(SGUESSMEMERR);
  end;

  try                              { use I/O buffers, finally release }
    try
      Size:=FInputStream.Size;
    except
      Raise ELzrw1Exception.Create(SGUESSSTRMSIZEERR);
    end;

    if FInputStream.Seek(0, soFromBeginning)<0 then
       Raise ELzrw1Exception.Create(SGUESSSTRMOPENERR);

    try                    { try to use the inputfile, finally close }

                           { small files can afford LZH }
      if (Size < (3 * ChunkSize)) then begin
        FCompressMode := Good;
        Result := Good;
        exit;
      end;
                           { try 2 blocks with fast at 1/3 and 2/3 of file }
      FInputStream.Seek((Size div 3) and $7FFF8000, soFromBeginning);

      SrcSize := FInputStream.Read(SrcBuf^, ChunkSize);
      if (SrcSize < 0) then begin
        Raise ELzrw1Exception.Create(SGUESSSTRMRDERR);
      end;
      UncompressedSize := SrcSize;
      CompressedSize := FLZR.Compression(SRCBuf,DSTBuf,SrcSize);

      FInputStream.Seek(((Size * 2)div 3) and $7FFF8000, soFromBeginning);
      SrcSize := FInputStream.Read(SrcBuf^, ChunkSize);
      if (SrcSize < 0) then begin
        Raise ELzrw1Exception.Create(SGUESSSTRMRDERR);
      end;
      Inc(UncompressedSize, SrcSize);
      Inc(CompressedSize,FLZR.Compression(SRCBuf,DSTBuf,SRCSize));

      if (((UnCompressedSize * FThreshold) div 100) > CompressedSize) then begin
        Result := Fast;
      end
      else begin
        Result := Good;
      end;

    finally
      InputStream.Seek(SavedStreamPos, soFromBeginning)
    end;

  finally
    Freemem(SRCBuf,IOBufSize);
    Freemem(DSTBuf,IOBufSize);
  end;

end;

end.
