{ VERSION: 1.5 }
{ 1.4 / 1.10.2005
  - opertor 'db.operator.strcat'
  1.3 / 15.5.2005
  - pepojeno na zkladn konektor TMyskaBaseJetStore
  - optimalizace vyuvajc exkluzivnho pstupu k databzi
  - implementuje detektor formtu souboru ValidFile()
  1.2 / 2.4.2005
  - pouziva specializovany Ado konektor s podporou komentaru v dotazech
  1.1 / 10.3.2005
  - definuje datove typu db.type.string,generator,date,time,datetime,memo,blob,double,money,integer,boolean
}
// TODO: implicitne nerozlisuji textova pole mala/velka pismena !! chyba
unit uMyskaJetStore;

interface
uses
  Classes, DB, uMyskaDB, uMyskaBaseJetStore;

type
  TMyskaJetStore = class(TMyskaBaseJetStore)
  protected
    procedure ConfigureQueries; override;
    procedure DoPrepareGenerators; override;
    function DoPrepareGenerator(const GeneratorName: string): string; override;
    procedure DoAdd(Sender: TMyskaApp); override;
  protected
    function DoLock(const LockName: string): TObject; override;
    procedure DoBeginUpdate; override;
    procedure DoEndUpdate; override;
    function ExclusiveAccess: Boolean; override;
  protected
    class function ValidFile(const Header: string): Boolean; override;
  protected
    function CmdGeneratorIncrement(const Params: Variant): TDataSet;
  end;

implementation
uses
  SysUtils;

class function TMyskaJetStore.ValidFile(const Header: string): Boolean;
const
  Hdr = 'Standard Jet DB';
begin
  Result := Copy(Header, 5, Length(Hdr)) = Hdr
end;

procedure TMyskaJetStore.DoAdd(Sender: TMyskaApp);
begin
  inherited;
  with Sender.DBQueries do begin
    ArrayValues['app.table.name', 1] := '{app.id}_%s';
    ArrayValues['app.table.name', 2] := '%0:s_%1:s';
    ArrayValues['app.generator.name', 1] := '{app.id}.%s';
  end
end;

function TMyskaJetStore.DoLock(const LockName: string): TObject;
begin
  // protoze mame vyhradni pristup neni to k nicemu potreba
  Result := nil
end;

procedure TMyskaJetStore.DoBeginUpdate;
begin
  // protoze mame vyhradni pristup neni to k nicemu potreba
end;

procedure TMyskaJetStore.DoEndUpdate;
begin
  // protoze mame vyhradni pristup neni to k nicemu potreba
end;

function TMyskaJetStore.ExclusiveAccess: Boolean;
begin
  Result := True
end;

procedure TMyskaJetStore.ConfigureQueries;
begin
  inherited;
  RegisterCommand('db.table.exists', CmdTableExists);
  RegisterCommand('db.generator.increment', CmdGeneratorIncrement);

  with FDBQueries do begin
    Values['db.operator.upper'] := 'UCase';
    ArrayValues['db.operator.strcat', 2] := '(CStr(%0:s) + CStr(%1:s))';
    ArrayValues['db.operator.strcomp', 2] := '(StrComp(%s,%s,0)=0)';
    ArrayValues['db.operator.stricomp', 2] := '(%s=%s)';
    ArrayValues['db.operator.istrue', 1] := '%s';
    Values['db.const.true'] := 'True';
    Values['db.const.false'] := 'False';
    ArrayValues['db.type.string', 1] := 'VARCHAR(%s)'; // COLLATE SQL_Czech_Cp1250_CS_AS';
    Values['db.type.generator'] := 'INTEGER';
    Values['db.type.date'] := 'DATE';
    Values['db.type.time'] := 'TIME';
    Values['db.type.datetime'] := 'TIMESTAMP';
    Values['db.type.memo'] := 'MEMO';
    Values['db.type.blob'] := 'LONGBINARY';
    Values['db.type.double'] := 'DOUBLE';
    Values['db.type.money'] := 'MONEY';
    Values['db.type.integer'] := 'INTEGER';
    // TODO:
    ArrayValues['db.type.boolean', 0] := 'BIT';  // 'BOOLEAN'
// TODO: ODBC ovladac to nesezere
//    ArrayValues['db.type.boolean', 1] := 'BIT DEFAULT %s'; // TODO: 'BOOLEAN DEFAULT %s';
    ArrayValues['db.type.boolean', 1] := 'BIT';

    ArrayValues['db.table.exists', 1] := 'db.table.exists(''%s'')';
    ArrayValues['db.table.drop', 1] := 'DROP TABLE %s';
    ArrayValues['db.generator.exists', 1] := 'SELECT GENERATORNAME FROM {mys.table.name(GENERATORS)} WHERE GENERATORNAME = {db.operator.upper}(''%s'')';
    ArrayValues['db.generator.create', 1] := 'INSERT INTO {mys.table.name(GENERATORS)}(GENERATORNAME, GENERATORVALUE) VALUES ({db.operator.upper}(''%s''), 0)';
    ArrayValues['db.generator.drop', 1] := 'DELETE FROM {mys.table.name(GENERATORS)} WHERE GENERATORNAME = {db.operator.upper}(''%s'')';
    ArrayValues['db.generator.value', 1] := 'SELECT GENERATORVALUE FROM {mys.table.name(GENERATORS)} WHERE GENERATORNAME = {db.operator.upper}(''%s'')';
    ArrayValues['db.generator.setValue', 2] := 'UPDATE {mys.table.name(GENERATORS)} SET GENERATORVALUE=%1:d WHERE GENERATORNAME = {db.operator.upper}(''%0:s'')';
    ArrayValues['db.generator.incrementValue', 1] := 'UPDATE {mys.table.name(GENERATORS)} SET GENERATORVALUE=GENERATORVALUE+1 WHERE GENERATORNAME = {db.operator.upper}(''%s'')';
    ArrayValues['db.generator.increment', 1] := 'db.generator.increment(''%s'')';

    ArrayValues['mys.table.name', 1] := 'MYS_%s';
  end
end;

procedure TMyskaJetStore.DoPrepareGenerators;
begin
  inherited;
  CreateTable(
    '{mys.table.name(GENERATORS)}',
      'GENERATORNAME {db.type.string(40)} NOT NULL PRIMARY KEY,' +
      'GENERATORVALUE {db.type.generator}'
  )
end;

const
  MaxGeneratorName = 30;     // TUNE:

function TMyskaJetStore.DoPrepareGenerator(const GeneratorName: string): string;
begin
  Result := UpperCase(GeneratorName);
  if Length(Result) > MaxGeneratorName then
    SetLength(Result, MaxGeneratorName);
  if not Exists('{db.generator.exists}', [Result]) then
    Exec('{db.generator.create}', [so_Command], [Result])
end;

function TMyskaJetStore.CmdGeneratorIncrement(const Params: Variant): TDataSet;
var
  Lck: TObject;
  GeneratorName: string;
begin
  GeneratorName := Params[0];
  Lck := Lock('GENID.' + GeneratorName);
  try
    Exec('{db.generator.incrementValue}', [so_Command], [GeneratorName]);
    Result := Exec('{db.generator.value}', [so_ReadOnly, so_ReadOnce], [GeneratorName]);
  finally
    Lck.Free
  end
end;

begin
  RegisterFactory(TMyskaJetStore, ['mdb'])
end.
