{ VERSION: 1.8 }
{ Definuje strom rznorodch objekt }
{ 1.7 / 7.6.2005
  - nov vlastnost TColumnDescriptor.Template: TColumnDescriptor ukazuje
    na deskriptor ze kterho byl proveden Assign()    
  1.6 / 12.5.2005
  - odstrann jeden memory leak
  1.5 / 11.5.2005
  - nov vlastnost TMappedColumn.Name udv intern jmno sloupce vhodn
    jako vazebn kl. Pokud nen dn Name piazen, tak vrac pseudo jmno
    generovan z TMappedColumn.Caption
  1.4 / 10.5.2005
  - kompatibiln s Delphi6
  1.3 / 9.5.2005
  - nov vlastnost TColumnDescriptor.Alignment: TColumnAlignment pro nastaven
    zarovnn pi zobrazen
  - formtovn hodnot na textov podoby TTreeNode.Strings[] prochz pes
    TColumnDescriptor.GetDisplayText()
  - formtovn relnch sel a datum se d vlastnostmi
    TColumnDescriptor.FloatFormat a TColumnDescriptor.DateFormat
  - hodnoty typu 'string' a 'integer' je mon pevdt pomoc seznamu
    TColumnDescriptor.ListFormat.
    Nap. ListFormat := '0=nula;1=jednika;4=tyka'
  - TColumnDescriptor podporuje operaci Assign(:TColumnDescriptor)
  1.2 / 2.5.2005
  - potomci mou pedefinovat adc funkci sloupce 
    ( viz. TColumnDescriptor.Compare() ). vchoz azen
    porovnv sla s sly a datumy s datumy. Kdy na sebe 
    datov typy nepasuj, tak se porovnaj jejich textov zpisy
  1.1 / 26.4.2005
}
unit uObjectTree;

interface
uses
  Classes;

const
  cdFloatFormat = 2;
  cdDateFormat = 3;
  cdListFormat = 4;
type
  TNodeKey = Pointer;
  TTreeNode = class;

  TColumnDescriptors = class;
  TColumnAlignment = (taDefault, taLeftJustify, taRightJustify, taCenter);
  TColumnDescriptor = class(TCollectionItem)
  private
    FTag: Integer;
    FTemplate: TColumnDescriptor;
    FAlignment: TColumnAlignment;
    FListNames: TStrings;
    FListValues: TStrings;
    FFormatAssigned: Boolean;
    FStrings: array [0..4] of string;
    function GetString(Index: Integer): string;
    procedure SetString(Index: Integer; const Value: string);
    procedure SetName(Index: Integer; const Value: string);
    function GetCollection: TColumnDescriptors;
  protected
    function GetListValue(const ID: string; var Value: string): Boolean; overload;
    function GetListValue(const ID: Integer; var Value: string): Boolean; overload;
    property FormatAssigned: Boolean read FFormatAssigned;
  public
    property Alignment: TColumnAlignment read FAlignment write FAlignment;
    property FloatFormat: string index cdFloatFormat read GetString write SetString;
    property DateFormat: string index cdDateFormat read GetString write SetString;
    property ListFormat: string index cdListFormat read GetString write SetString;
  public
    function Compare(Node1, Node2: TTreeNode): Integer; virtual;
    function GetDisplayText(Node: TTreeNode): string; virtual;
    function GetAlignment(Node: TTreeNode): TAlignment; virtual;
  public
    property Name: string index 0 read GetString write SetName;
    property Caption: string index 1 read GetString write SetString;
    property Collection: TColumnDescriptors read GetCollection;
    property Index;
  public
    procedure Assign(Source: TPersistent); override;
    destructor Destroy; override;
    property Template: TColumnDescriptor read FTemplate;
    property Tag: Integer read FTag write FTag;
  end;

  TNodeDescriptor = class;
  TColumnDescriptorClass = class of TColumnDescriptor;
  TColumnDescriptors = class(TCollection)
  private
    FDescriptor: TNodeDescriptor;
    function GetItem(Index: Integer): TColumnDescriptor;
  public
    constructor Create(ItemClass: TColumnDescriptorClass);
    function Add: TColumnDescriptor;
    function IndexOf(const Name: string): Integer;
    property Items[Index: Integer]: TColumnDescriptor read GetItem; default;
    property Descriptor: TNodeDescriptor read FDescriptor;
  end;

  TNodeDescriptors = class;
  TNodeDescriptorList = class(TList)
  private
    function GetItem(Index: Integer): TNodeDescriptor;
  public
    function Add(Item: TNodeDescriptor; Recurse: Boolean = False): Integer; overload;
    function Add(Node: TTreeNode): Integer; overload;
    function Add(List: TNodeDescriptorList): Integer; overload;
    function Add(Descriptors: TNodeDescriptors): Integer; overload;
    function IndexOf(Item: TNodeDescriptor): Integer;
    property Count;
    property Items[Index: Integer]: TNodeDescriptor read GetItem; default;
  end;
  TNodeDescriptor = class(TCollectionItem)
  private
    FDescriptors: TNodeDescriptorList;
    FName: string;
    FColumns: TColumnDescriptors;
    function GetColumns: TColumnDescriptors;
    function GetCollection: TNodeDescriptors;
    procedure SetName(const Value: string);
  protected
    function CreateColumns: TColumnDescriptors; virtual;
  public
    procedure AfterConstruction; override;
    destructor Destroy; override;
    property Collection: TNodeDescriptors read GetCollection;
    property Name: string read FName write SetName;
    property Index;
    property Columns: TColumnDescriptors read GetColumns;
    property Descriptors: TNodeDescriptorList read FDescriptors;
  end;

  TNodeDescriptorClass = class of TNodeDescriptor;
  TNodeDescriptors = class(TCollection)
  private
    function GetItem(Index: Integer): TNodeDescriptor;
  public
    constructor Create(ItemClass: TNodeDescriptorClass);
    function Add: TNodeDescriptor;
    function IndexOf(const Name: string): Integer;
    property Items[Index: Integer]: TNodeDescriptor read GetItem; default;
  end;

  TMappedColumns = class;
  TMappedColumn = class(TCollectionItem)
  private
    FName: string;
    FCaption: string;
    FColumns: array of TColumnDescriptor;
    function GetCollection: TMappedColumns;
    function GetName: string;
  protected
    procedure Add(Column: TColumnDescriptor);
    function GetColumn(Index: Integer): TColumnDescriptor;
    function IndexOf(Column: TColumnDescriptor): Integer;
  public
    function Column(Descriptor: TNodeDescriptor): TColumnDescriptor; overload;
    property Collection: TMappedColumns read GetCollection;
    property Name: string read GetName write FName;
    property Caption: string read FCaption write FCaption;
  public
    function ColumnCount: Integer;
    property Columns[Index: Integer]: TColumnDescriptor read GetColumn; default;
  end;

  TMappedColumns = class(TCollection)
  private
    function GetItem(Index: Integer): TMappedColumn;
  protected
    function Add: TMappedColumn;
  public
    constructor Create;
    function IndexOf(Column: TColumnDescriptor): Integer; overload;
    function IndexOf(const Name: string): Integer; overload;
    property Items[Index: Integer]: TMappedColumn read GetItem; default;
    property Count;
  end;

  TNodeManager = class
  private
    FDescriptors: TNodeDescriptors;
    function GetDescriptors: TNodeDescriptors;
  protected
    function CreateDescriptors: TNodeDescriptors; virtual;
    function ColumnsEqual(A: TColumnDescriptor; B: TColumnDescriptor; Context: TTreeNode): Boolean; virtual;
  public
    destructor Destroy; override;
    function CreateColumnMap(Node: TTreeNode): TMappedColumns; virtual;
    function CreateRootNode(ADescriptor: TNodeDescriptor): TTreeNode; virtual; abstract;
    property Descriptors: TNodeDescriptors read GetDescriptors;
  end;

  TTree = class
  private
    FRoot: TTreeNode;
    FConnected: Boolean;
    FManager: TNodeManager;
    FColumns: TMappedColumns;
    function GetColumns: TMappedColumns;
    procedure SetColumns(Value: TMappedColumns);
    procedure SetConnected(Value: Boolean);
    procedure SetRoot(Value: TTreeNode);
  protected
    procedure DoConnect; virtual;
    procedure DoDisconnect; virtual;
  public
//    procedure Expand(Value: TTreeNode);
    procedure AddChild(Parent: TTreeNode; Value: TTreeNode); overload; virtual; abstract;
    procedure AddChild(Value: TTreeNode); overload;
  public
    destructor Destroy; override;
    property Root: TTreeNode read FRoot write SetRoot;
    property Manager: TNodeManager read FManager;
    property Columns: TMappedColumns read GetColumns write SetColumns;
    property Connected: Boolean read FConnected write SetConnected;
  end;

  TNodeView = Pointer;
  TTreeNode = class
  private
    FView: TNodeView;
  protected
    function GetTree: TTree; virtual;
    function GetManager: TNodeManager; virtual;
    function GetDescriptor: TNodeDescriptor; virtual;
    function GetKey: TNodeKey; virtual;
    procedure Locate(const Value: TNodeKey); virtual;
  protected
    function GetValue(Index: Integer): Variant; virtual;
    procedure SetValue(Index: Integer; const Value: Variant); virtual;
    function GetValueAsString(Index: Integer): string; virtual;
    procedure SetValueAsString(Index: Integer; const Value: string); virtual;
    function GetValueAsCode(Index: Integer): string; virtual;
    procedure SetValueAsCode(Index: Integer; const Value: string); virtual;
  public
//    function HasChildren: Boolean; virtual;
    function ColumnIndex(const Name: string): Integer;
    function ColumnName(Index: Integer): string;

    function CreateNodeDescriptorList: TNodeDescriptorList; virtual;

    property View: TNodeView read FView write FView;
    property Key: TNodeKey read GetKey write Locate;
    property Tree: TTree read GetTree;
    property Manager: TNodeManager read GetManager;
    property Descriptor: TNodeDescriptor read GetDescriptor;

    property Values[Index: Integer]: Variant read GetValue write SetValue;
    property Strings[Index: Integer]: string read GetValueAsString write SetValueAsString; default;
    property Codes[Index: Integer]: string read GetValueAsCode write SetValueAsCode;
  end;

function EncodeValue(const Value: Variant): string;
function DecodeValue(const Value: string): Variant;

function NormalizedName(const Value: string): string;

function VarCompare(const V1, V2: Variant): Integer;

implementation
uses
  SysUtils, uPropSet, Variants;

function VarCompare(const V1, V2: Variant): Integer;
begin
  if not VarIsEmpty(V1) and not VarIsEmpty(V2) then
  try
    if V1 = V2 then
      Result := 0
    else
      if V1 < V2 then
        Result := -1
      else
        Result := 1;
  except
    on EVariantError do
      Result := VarCompare(string(V1), string(V2));
  end
  else
    Result := VarCompare(string(V1), string(V2));
end;

function EncodeValue(const Value: Variant): string;
begin
  Result := uPropSet.EncodeVariant(Value)
end;

function DecodeValue(const Value: string): Variant;
begin
  Result := uPropSet.DecodeVariant(Value)
end;

function NormalizedName(const Value: string): string;
var
  I: Integer;
begin
  Result := Value;
  for I := 1 to Length(Result) do
    Result[I] := UpCase(Result[I])
end;

// TMappedColumn -----------------------------------------------------------------

function TMappedColumn.GetCollection: TMappedColumns;
begin
  Result := TMappedColumns(inherited Collection)
end;

function PlainText(const S: string): string;
const
  // TODO: podpo i jin jazyky
  Ups  = '؊ݎ';
  Lows = 'ACDEEINORSTUYZ';
var
  I, J: Integer;
begin
  Result := AnsiUpperCase(S);
  for I := 1 to Length(Result) do begin
    J := Pos(Result[I], Ups);
    if J <> 0 then
      Result[I] := Lows[J];
    if not (Result[I] in ['A'..'Z', '0'..'9']) then
      Result[I] := '_'
  end
end;

function TMappedColumn.GetName: string;
begin
  Result := FName;
  if Result = '' then begin
    Result := PlainText(FCaption);
    if Result = '' then
      Result := Format('COL%d', [Index + 1]);
  end
end;

procedure TMappedColumn.Add(Column: TColumnDescriptor);
var
  D: TNodeDescriptor;
  I: Integer;
begin
  if Assigned(Column) then begin
    D := Column.Collection.Descriptor;
    if Assigned(D) then begin
      I := D.Index;
      if I >= 0 then begin
        if I >= Length(FColumns) then
          SetLength(FColumns, I + 1);
        FColumns[I] := Column;
        // vychozi popisek se vezme z prvniho popisku co prijde
        if Caption = '' then
          Caption := Column.Caption
      end
    end;
  end
end;

function TMappedColumn.Column(Descriptor: TNodeDescriptor): TColumnDescriptor;
var
  I: Integer;
begin
  Result := nil;
  if Assigned(Descriptor) then begin
    I := Descriptor.Index;
    if (I >= 0) and (I < Length(FColumns)) then
      Result := FColumns[I]
  end
end;

function TMappedColumn.ColumnCount: Integer;
var
  I: Integer;
begin
  Result := 0;
  for I := 0 to Length(FColumns) - 1 do
    if Assigned(FColumns[I]) then
      Inc(Result)
end;

function TMappedColumn.GetColumn(Index: Integer): TColumnDescriptor;
var
  I: Integer;
begin
  for I := 0 to Length(FColumns) - 1 do
    if Assigned(FColumns[I]) then
      if Index = 0 then begin
        Result := FColumns[I];
        exit
      end else
        Dec(Index);
  Result := nil
end;

function TMappedColumn.IndexOf(Column: TColumnDescriptor): Integer;
var
  I: Integer;
  Index: Integer;
begin
  Index := -1;
  for I := 0 to Length(FColumns) - 1 do
    if Assigned(FColumns[I]) then begin
      Inc(Index);
      if FColumns[I] = Column then begin
        Result := Index;
        exit
      end
    end;
  Result := -1
end;

// TMappedColumns -------------------------------------------------------------

constructor TMappedColumns.Create;
begin
  inherited Create(TMappedColumn)
end;

function TMappedColumns.Add: TMappedColumn;
begin
  Result := TMappedColumn(inherited Add)
end;

function TMappedColumns.GetItem(Index: Integer): TMappedColumn;
begin
  Result := TMappedColumn(inherited Items[Index])
end;

function TMappedColumns.IndexOf(Column: TColumnDescriptor): Integer;
var
  I: Integer;
  Col: TMappedColumn;
begin
  Result := -1;
  for I := 0 to Count - 1 do begin
    Col := Items[I];
    if Col.IndexOf(Column) >= 0 then begin
      Result := I;
      exit
    end
  end
end;

function TMappedColumns.IndexOf(const Name: string): Integer;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    if SameText(Items[I].Name, Name) then begin
      Result := I;
      exit
    end;
  Result := -1
end;

// TColumnDescriptor ----------------------------------------------------------

destructor TColumnDescriptor.Destroy;
begin
  FListNames.Free;
  FListValues.Free;
  inherited
end;

function TColumnDescriptor.GetCollection: TColumnDescriptors;
begin
  Result := TColumnDescriptors(inherited Collection)
end;

function TColumnDescriptor.GetString(Index: Integer): string;
begin
  Result := FStrings[Index]
end;

procedure TColumnDescriptor.SetString(Index: Integer; const Value: string);
begin
  if Value <> FStrings[Index] then begin
    FStrings[Index] := Value;
    if Index in [cdFloatFormat, cdDateFormat, cdListFormat] then begin
      FFormatAssigned := (FStrings[cdFloatFormat] <> '') or (FStrings[cdDateFormat] <> '') or (FStrings[cdListFormat] <> '');
      if Index = cdListFormat then begin
        FreeAndNil(FListNames);
        FreeAndNil(FListValues);
      end
    end;
    // TODO: Changed
  end
end;

procedure TColumnDescriptor.SetName(Index: Integer; const Value: string);
begin
  SetString(Index, NormalizedName(Value))
end;

function TColumnDescriptor.Compare(Node1, Node2: TTreeNode): Integer;
begin
  Result := VarCompare(Node1.Values[Index], Node2.Values[Index])
end;

{$IFDEF VER140}
type
  TStringList = class(Classes.TStringList)
  private
    function GetValueFromIndex(Index: Integer): string;
  public
    property ValueFromIndex[Index: Integer]: string read GetValueFromIndex;
  end;

function TStringList.GetValueFromIndex(Index: Integer): string;
var
  I: Integer;
begin
  Result := Strings[Index];
  I := Pos('=', Result);
  if I = 0 then
    Result := ''
  else
    System.Delete(Result, 1, I)
end;
{$ENDIF}

function TColumnDescriptor.GetListValue(const ID: string; var Value: string): Boolean;
var
  List: TStringList;
  Key: string;
  KeyValue: string;
  I: Integer;
begin
  if not Assigned(FListNames) then begin
    if ListFormat = '' then begin
      Value := ID;
      Result := False;
      exit
    end;
    FListNames := TStringList.Create;
    TStringList(FListNames).Sorted := True;
    FListValues := TStringList.Create;

    List := TStringList.Create;
    try
      List.Text := StringReplace(StringReplace(ListFormat, ',' { TUNE: }, #13#10, [rfReplaceAll]), ';' { TUNE: }, #13#10, [rfReplaceAll]);
      for I := 0 to List.Count - 1 do begin
        Key := List.Names[I];
        KeyValue := List.ValueFromIndex[I];
        FListNames.AddObject(Key, TObject(FListValues.Add(KeyValue)))
      end
    finally
      List.Free
    end
  end;
  I := FListNames.IndexOf(ID);
  if I < 0 then begin
    Value := ID;
    Result := False
  end else begin
    Value := FListValues[Integer(FListNames.Objects[I])];
    Result := True
  end
end;

function TColumnDescriptor.GetListValue(const ID: Integer; var Value: string): Boolean;
begin
  Result := GetListValue(IntToStr(ID), Value)
end;

function TColumnDescriptor.GetDisplayText(Node: TTreeNode): string;
var
  Code: string;
label
  defaultFormat;
begin
  if FormatAssigned then begin
    Code := Node.Codes[Index];
    case DecodeVarType(Code) of
    varDate:
      if DateFormat <> '' then
        Result := FormatDateTime(DateFormat, DecodeValue(Code))
      else
        goto defaultFormat;

    varDouble:
      if FloatFormat <> '' then
        Result := FormatFloat(FloatFormat, DecodeValue(Code))
      else
        goto defaultFormat;

    varString:
      GetListValue(string(DecodeValue(Code)), Result);

    varInteger:
      GetListValue(Integer(DecodeValue(Code)), Result);

    else
      defaultFormat:
      Result := DecodeValue(Code)
    end
  end else
    Result := Node.Values[Index]
end;

function TColumnDescriptor.GetAlignment(Node: TTreeNode): TAlignment;
begin
  if Alignment = taDefault then begin
    case DecodeVarType(Node.Codes[Index]) of
    varInteger,
    varDate,
    varDouble:
      Result := Classes.taRightJustify;
    else
      Result := Classes.taLeftJustify
    end;
  end else
    Result := TAlignment(Ord(Alignment) - Ord(taLeftJustify))
end;

procedure TColumnDescriptor.Assign(Source: TPersistent);
var
  Col: TColumnDescriptor absolute Source;
begin
  if Source is TColumnDescriptor then begin
    Alignment := Col.Alignment;
    FloatFormat := Col.FloatFormat;
    DateFormat := Col.DateFormat;
    ListFormat := Col.ListFormat;
    Caption := Col.Caption;
    FTemplate := Col
  end else
    inherited
end;

// TColumnDescriptors ---------------------------------------------------------

function TColumnDescriptors.GetItem(Index: Integer): TColumnDescriptor;
begin
  Result := TColumnDescriptor(inherited Items[Index])
end;

function TColumnDescriptors.Add: TColumnDescriptor;
begin
  Result := TColumnDescriptor(inherited Add)
end;

constructor TColumnDescriptors.Create(ItemClass: TColumnDescriptorClass);
begin
  inherited Create(ItemClass)
end;

function TColumnDescriptors.IndexOf(const Name: string): Integer;
var
  I: Integer;
  S: string;
begin
  S := NormalizedName(Name);
  for I := Count - 1 downto 0 do
    if Items[I].Name = S then begin
      Result := I;
      exit
    end;
  Result := -1
end;

// TNodeDescriptors -----------------------------------------------------------

function TNodeDescriptors.GetItem(Index: Integer): TNodeDescriptor;
begin
  Result := TNodeDescriptor(inherited Items[Index])
end;

constructor TNodeDescriptors.Create(ItemClass: TNodeDescriptorClass);
begin
  inherited Create(ItemClass)
end;

function TNodeDescriptors.Add: TNodeDescriptor;
begin
  Result := TNodeDescriptor(inherited Add)
end;

function TNodeDescriptors.IndexOf(const Name: string): Integer;
var
  I: Integer;
  S: string;
begin
  S := NormalizedName(Name);
  for I := Count - 1 downto 0 do
    if Items[I].Name = S then begin
      Result := I;
      exit
    end;
  Result := -1
end;

// TNodeDescriptor ------------------------------------------------------------

function TNodeDescriptor.CreateColumns: TColumnDescriptors;
begin
  Result := TColumnDescriptors.Create(TColumnDescriptor)
end;

function TNodeDescriptor.GetColumns: TColumnDescriptors;
begin
  Result := FColumns;
  if not Assigned(Result) then begin
    FColumns := CreateColumns;
    if Assigned(FColumns) then begin
      FColumns.FDescriptor := Self;
      Result := FColumns
    end
  end
end;

procedure TNodeDescriptor.AfterConstruction;
begin
  FDescriptors := TNodeDescriptorList.Create;
  inherited
end;

destructor TNodeDescriptor.Destroy;
begin
  inherited;
  FColumns.Free;
  FDescriptors.Free
end;

function TNodeDescriptor.GetCollection: TNodeDescriptors;
begin
  Result := TNodeDescriptors(inherited Collection)
end;

procedure TNodeDescriptor.SetName(const Value: string);
begin
  FName := NormalizedName(Value)
end;

// TNodeDescriptorList --------------------------------------------------------

function TNodeDescriptorList.Add(Descriptors: TNodeDescriptors): Integer;
var
  I: Integer;
begin
  Result := -1;
  if Assigned(Descriptors) then
    for I := 0 to Descriptors.Count - 1 do
      Add(Descriptors[I])
end;

function TNodeDescriptorList.Add(List: TNodeDescriptorList): Integer;
var
  I: Integer;
begin
  Result := -1;
  if Assigned(List) then
    for I := 0 to List.Count - 1 do
      Add(List[I])
end;

function TNodeDescriptorList.Add(Item: TNodeDescriptor; Recurse: Boolean): Integer;
var
  R: TNodeDescriptorList;

  procedure A(D: TNodeDescriptor);
  var
    L: TNodeDescriptorList;
    I: Integer;
  begin
    if not Assigned(D) then exit;
    R.Add(D);
    L := D.Descriptors;
    for I := 0 to L.Count - 1 do begin
      D := L[I];
      if R.IndexOf(D) < 0 then
        A(D)
    end
  end;

begin
  if Assigned(Item) then begin
    if Recurse then begin
      R := TNodeDescriptorList.Create;
      try
        A(Item);
        Result := Add(R)
      finally
        R.Free
      end
    end else begin
      Result := IndexOf(Item);
      if Result < 0 then
        Result := inherited Add(Item)
    end
  end else
    Result := -1
end;

function TNodeDescriptorList.Add(Node: TTreeNode): Integer;
var
  L: TNodeDescriptorList;
begin
  if Assigned(Node) then begin
    L := Node.CreateNodeDescriptorList;
    try
      Result := Add(L)
    finally
      L.Free
    end
  end else
    Result := -1
end;

function TNodeDescriptorList.IndexOf(Item: TNodeDescriptor): Integer;
begin
  Result := inherited IndexOf(Item)
end;

function TNodeDescriptorList.GetItem(Index: Integer): TNodeDescriptor;
begin
  Result := TNodeDescriptor(inherited Items[Index])
end;

// TNodeManager ---------------------------------------------------------------

function TNodeManager.CreateColumnMap(Node: TTreeNode): TMappedColumns;
var
  M: TMappedColumns;

  procedure Add(Columns: TColumnDescriptors);

    procedure Add(Column: TColumnDescriptor);
    var
      I, J: Integer;
      C: TMappedColumn;
      Col: TColumnDescriptor;
    begin
      for I := 0 to M.Count - 1 do begin
        C := M[I];
        for J := 0 to C.ColumnCount - 1 do begin
          Col := C[J];
          if (Column.Collection <> Col.Collection) and ColumnsEqual(Column, Col, Node) then begin
            C.Add(Column);
            exit
          end;
        end
      end;
      C := M.Add;
      C.Add(Column);
    end;

  var
    I: Integer;
  begin
    for I := 0 to Columns.Count - 1 do
      Add(Columns[I]);
  end;

var
  L: TNodeDescriptorList;
  I: Integer;
begin
  M := TMappedColumns.Create;
  Result := M;
  if not Assigned(Node) then exit;
  L := Node.CreateNodeDescriptorList;
  try
    for I := 0 to L.Count - 1 do
      Add(L[I].Columns)
  finally
    L.Free
  end
end;

function TNodeManager.ColumnsEqual(A: TColumnDescriptor; B: TColumnDescriptor; Context: TTreeNode): Boolean;
begin
  Result := ((A.Name = B.Name) and (A.Name <> '')) or
            ((A.Caption = B.Caption) and (A.Caption <> ''))
end;

function TNodeManager.GetDescriptors: TNodeDescriptors;
begin
  Result := FDescriptors;
  if not Assigned(Result) then begin
    FDescriptors := CreateDescriptors;
    Result := FDescriptors
  end
end;

function TNodeManager.CreateDescriptors: TNodeDescriptors;
begin
  Result := TNodeDescriptors.Create(TNodeDescriptor)
end;

destructor TNodeManager.Destroy;
begin
  inherited;
  FDescriptors.Free
end;

// TTreeNode ------------------------------------------------------------------

function TTreeNode.GetKey: TNodeKey;
begin
  Result := nil
end;

procedure TTreeNode.Locate(const Value: TNodeKey);
begin
end;

function TTreeNode.GetTree: TTree;
begin
  Result := nil
end;

function TTreeNode.GetManager: TNodeManager;
begin
  Result := nil
end;

function TTreeNode.GetDescriptor: TNodeDescriptor;
begin
  Result := nil
end;

function TTreeNode.ColumnIndex(const Name: string): Integer;
var
  D: TNodeDescriptor;
begin
  D := Descriptor;
  if Assigned(D) then
    Result := D.Columns.IndexOf(Name)
  else
    Result := -1
end;

function TTreeNode.ColumnName(Index: Integer): string;
var
  D: TNodeDescriptor;
begin
  D := Descriptor;
  if Assigned(D) and (Index >= 0) and (Index < D.Columns.Count) then
    Result := D.Columns[Index].Name
  else
    Result := ''
end;

function TTreeNode.GetValue(Index: Integer): Variant;
begin
  Result := DecodeValue(Codes[Index])
end;

procedure TTreeNode.SetValue(Index: Integer; const Value: Variant);
begin
  Codes[Index] := EncodeValue(Value)
end;

function TTreeNode.GetValueAsString(Index: Integer): string;
var
  D: TNodeDescriptor;
begin
  D := Descriptor;
  if Assigned(D) and (Index >= 0) and (Index < D.Columns.Count) then
    Result := D.Columns[Index].GetDisplayText(Self)
  else
    Result := Values[Index]
end;

procedure TTreeNode.SetValueAsString(Index: Integer; const Value: string);
begin
  Values[Index] := Value
end;

function TTreeNode.GetValueAsCode(Index: Integer): string;
begin
  Result := EncodeValue(Values[Index])
end;

procedure TTreeNode.SetValueAsCode(Index: Integer; const Value: string);
begin
  Values[Index] := DecodeValue(Value)
end;

function TTreeNode.CreateNodeDescriptorList: TNodeDescriptorList;
begin
  Result := TNodeDescriptorList.Create;
  Result.Add(Descriptor, True)
end;

// TTree ----------------------------------------------------------------------

procedure TTree.AddChild(Value: TTreeNode);
begin
  AddChild(nil, Value)
end;

function TTree.GetColumns: TMappedColumns;
begin
  Result := FColumns;
  if not Assigned(Result) then begin
    FColumns := TMappedColumns.Create;
    Result := FColumns
  end;
end;

procedure TTree.SetColumns(Value: TMappedColumns);
var
  OldConnected: Boolean;
begin
  if FColumns <> Value then begin
    OldConnected := Connected;
    Connected := False;
    try
      FColumns.Free;
      FColumns := Value;
    finally
      Connected := OldConnected
    end
  end
end;

procedure TTree.SetConnected(Value: Boolean);
begin
  if FConnected <> Value then begin
    if Value then begin
      if Assigned(FRoot) then begin
        DoConnect;
        FConnected := True
      end
    end else begin
      DoDisconnect;
      FConnected := False;
    end
  end;
end;

procedure TTree.SetRoot(Value: TTreeNode);
var
  OldConnected: Boolean;
begin
  if FRoot <> Value then begin
    OldConnected := Connected;
    Connected := False;
    try
      if Assigned(FRoot) then
        FRoot.Free;
      if Assigned(Value) then begin
        FRoot := Value;
        FManager := FRoot.Manager;
        if Assigned(FManager) then
          Columns := Manager.CreateColumnMap(Value)
      end else begin
        FRoot := nil;
        FColumns.Free;
        FColumns := nil;
        FManager := nil
      end;
    finally
      Connected := OldConnected
    end
  end
end;

destructor TTree.Destroy;
begin
  Connected := False;
  Root := nil;
  Columns := nil;
  inherited;
end;

procedure TTree.DoConnect;
begin
end;

procedure TTree.DoDisconnect;
begin
end;

end.
